from __future__ import annotations

import glob
import json
import os
import typing
from copy import deepcopy
from pathlib import Path
from typing import Any, Generic, cast

import cv2
import hydra
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import timm
import torch
from joblib import dump, load
from omegaconf import DictConfig, ListConfig, OmegaConf
from pytorch_grad_cam import GradCAM
from scipy import ndimage
from sklearn.base import ClassifierMixin
from sklearn.metrics import ConfusionMatrixDisplay
from torch import nn
from torchinfo import summary
from tqdm import tqdm

from quadra.callbacks.mlflow import get_mlflow_logger
from quadra.callbacks.scheduler import WarmupInit
from quadra.datamodules import (
    ClassificationDataModule,
    MultilabelClassificationDataModule,
    SklearnClassificationDataModule,
)
from quadra.datasets.classification import ImageClassificationListDataset
from quadra.models.base import ModelSignatureWrapper
from quadra.models.classification import BaseNetworkBuilder
from quadra.models.evaluation import BaseEvaluationModel, TorchEvaluationModel, TorchscriptEvaluationModel
from quadra.modules.classification import ClassificationModule
from quadra.tasks.base import Evaluation, LightningTask, Task
from quadra.trainers.classification import SklearnClassificationTrainer
from quadra.utils import utils
from quadra.utils.classification import (
    get_results,
    save_classification_result,
)
from quadra.utils.evaluation import automatic_datamodule_batch_size
from quadra.utils.export import export_model, import_deployment_model
from quadra.utils.models import get_feature, is_vision_transformer
from quadra.utils.vit_explainability import VitAttentionGradRollout

log = utils.get_logger(__name__)

SklearnClassificationDataModuleT = typing.TypeVar(
    "SklearnClassificationDataModuleT", bound=SklearnClassificationDataModule
)
ClassificationDataModuleT = typing.TypeVar("ClassificationDataModuleT", bound=ClassificationDataModule)


# TODO: Maybe we should have a BaseClassificationTask that is extended by Classification and MultilabelClassification
# at the current time, multilabel experiments use this Classification task class and they can not generate report
# (it is written specifically for a vanilla classification). Moreover, this class takes generic
# ClassificationDataModuleT, but multilabel experim. uses MultilabelClassificationDataModule, which is not a child of
# ClassificationDataModule
class Classification(Generic[ClassificationDataModuleT], LightningTask[ClassificationDataModuleT]):
    """Classification Task.

    Args:
        config: The experiment configuration
        output: The otuput configuration.
        gradcam: Whether to compute gradcams
        checkpoint_path: The path to the checkpoint to load the model from. Defaults to None.
        lr_multiplier: The multiplier for the backbone learning rate. Defaults to None.
        output: The ouput configuration (under task config). It contains the bool "example" to generate
            figs of discordant/concordant predictions.
        report: Whether to generate a report containing the results after test phase
        run_test: Whether to run the test phase.
    """

    def __init__(
        self,
        config: DictConfig,
        output: DictConfig,
        checkpoint_path: str | None = None,
        lr_multiplier: float | None = None,
        gradcam: bool = False,
        report: bool = False,
        run_test: bool = False,
    ):
        super().__init__(
            config=config,
            checkpoint_path=checkpoint_path,
            run_test=run_test,
            report=report,
        )
        self.output = output
        self.gradcam = gradcam
        self._lr_multiplier = lr_multiplier
        self._pre_classifier: nn.Module
        self._classifier: nn.Module
        self._model: nn.Module
        self._optimizer: torch.optim.Optimizer
        self._scheduler: torch.optim.lr_scheduler._LRScheduler
        self.model_json: dict[str, Any] | None = None
        self.export_folder: str = "deployment_model"
        self.deploy_info_file: str = "model.json"
        self.report_confmat: pd.DataFrame
        self.best_model_path: str | None = None

    @property
    def optimizer(self) -> torch.optim.Optimizer:
        """Get the optimizer."""
        return self._optimizer

    @optimizer.setter
    def optimizer(self, optimizer_config: DictConfig) -> None:
        """Set the optimizer."""
        if (
            isinstance(self.model.features_extractor, nn.Module)
            and isinstance(self.model.pre_classifier, nn.Module)
            and isinstance(self.model.classifier, nn.Module)
        ):
            log.info("Instantiating optimizer <%s>", self.config.optimizer["_target_"])
            if self._lr_multiplier is not None and self._lr_multiplier > 0:
                params = [
                    {
                        "params": self.model.features_extractor.parameters(),
                        "lr": optimizer_config.lr * self._lr_multiplier,
                    }
                ]
            else:
                params = [{"params": self.model.features_extractor.parameters(), "lr": optimizer_config.lr}]
            params.append({"params": self.model.pre_classifier.parameters(), "lr": optimizer_config.lr})
            params.append({"params": self.model.classifier.parameters(), "lr": optimizer_config.lr})
            self._optimizer = hydra.utils.instantiate(optimizer_config, params)

    @property
    def len_train_dataloader(self) -> int:
        """Get the length of the train dataloader."""
        len_train_dataloader = len(self.datamodule.train_dataloader())
        if self.devices is not None:
            num_gpus = len(self.devices) if isinstance(self.devices, list) else 1
            len_train_dataloader = len_train_dataloader // num_gpus
            if not self.datamodule.train_dataloader().drop_last:
                len_train_dataloader += int(len(self.datamodule.train_dataloader()) % num_gpus != 0)
        return len_train_dataloader

    @property
    def scheduler(self) -> torch.optim.lr_scheduler._LRScheduler:
        """Get the scheduler."""
        return self._scheduler

    @scheduler.setter
    def scheduler(self, scheduler_config: DictConfig) -> None:
        log.info("Instantiating scheduler <%s>", scheduler_config["_target_"])
        if "CosineAnnealingWithLinearWarmUp" in self.config.scheduler["_target_"]:
            # This scheduler will be overwritten by the SSLCallback
            self._scheduler = hydra.utils.instantiate(
                scheduler_config,
                optimizer=self.optimizer,
                batch_size=1,
                len_loader=1,
            )
            self.add_callback(WarmupInit(scheduler_config=scheduler_config))
        else:
            self._scheduler = hydra.utils.instantiate(scheduler_config, optimizer=self.optimizer)

    @property
    def module(self) -> ClassificationModule:
        """Get the module of the model."""
        return self._module

    @LightningTask.module.setter
    def module(self, module_config):  # noqa: F811
        """Set the module of the model."""
        module = hydra.utils.instantiate(
            module_config,
            model=self.model,
            optimizer=self.optimizer,
            lr_scheduler=self.scheduler,
            gradcam=self.gradcam,
        )
        if self.checkpoint_path is not None:
            log.info("Loading model from lightning checkpoint: %s", self.checkpoint_path)
            module = module.__class__.load_from_checkpoint(
                self.checkpoint_path,
                model=self.model,
                optimizer=self.optimizer,
                lr_scheduler=self.scheduler,
                criterion=module.criterion,
                gradcam=self.gradcam,
            )
        self._module = module

    @property
    def pre_classifier(self) -> nn.Module:
        return self._pre_classifier

    @pre_classifier.setter
    def pre_classifier(self, model_config: DictConfig) -> None:
        if "pre_classifier" in model_config and model_config.pre_classifier is not None:
            log.info("Instantiating pre_classifier <%s>", model_config.pre_classifier["_target_"])
            self._pre_classifier = hydra.utils.instantiate(model_config.pre_classifier, _convert_="partial")
        else:
            log.info("No pre-classifier found in config: instantiate a torch.nn.Identity instead")
            self._pre_classifier = nn.Identity()

    @property
    def classifier(self) -> nn.Module:
        return self._classifier

    @classifier.setter
    def classifier(self, model_config: DictConfig) -> None:
        if "classifier" in model_config:
            log.info("Instantiating classifier <%s>", model_config.classifier["_target_"])
            if self.datamodule.num_classes is None or self.datamodule.num_classes < 2:
                raise ValueError(f"Non compliant datamodule.num_classes : {self.datamodule.num_classes}")
            self._classifier = hydra.utils.instantiate(
                model_config.classifier, out_features=self.datamodule.num_classes, _convert_="partial"
            )
        else:
            raise ValueError("A `classifier` definition must be specified in the config")

    @property
    def model(self) -> nn.Module:
        return self._model

    @model.setter
    def model(self, model_config: DictConfig) -> None:
        self.pre_classifier = model_config  # type: ignore[assignment]
        self.classifier = model_config  # type: ignore[assignment]
        log.info("Instantiating backbone <%s>", model_config.model["_target_"])
        self._model = hydra.utils.instantiate(
            model_config.model, classifier=self.classifier, pre_classifier=self.pre_classifier, _convert_="partial"
        )
        if getattr(self.config.backbone, "freeze_parameters_name", None) is not None:
            self.freeze_layers_by_name(self.config.backbone.freeze_parameters_name)

        if getattr(self.config.backbone, "freeze_parameters_index", None) is not None:
            frozen_parameters_indices: list[int]
            if isinstance(self.config.backbone.freeze_parameters_index, int):
                # Freeze all layers up to the specified index
                frozen_parameters_indices = list(range(self.config.backbone.freeze_parameters_index + 1))
            elif isinstance(self.config.backbone.freeze_parameters_index, ListConfig):
                frozen_parameters_indices = cast(
                    list[int], OmegaConf.to_container(self.config.backbone.freeze_parameters_index, resolve=True)
                )
            else:
                raise ValueError("freeze_parameters_index must be an int or a list of int")

            self.freeze_parameters_by_index(frozen_parameters_indices)

    def prepare(self) -> None:
        """Prepare the experiment."""
        super().prepare()
        self.model = self.config.model
        self.optimizer = self.config.optimizer
        self.scheduler = self.config.scheduler
        self.module = self.config.model.module

    def train(self):
        """Train the model."""
        super().train()
        if (
            self.trainer.checkpoint_callback is not None
            and hasattr(self.trainer.checkpoint_callback, "best_model_path")
            and self.trainer.checkpoint_callback.best_model_path is not None
            and len(self.trainer.checkpoint_callback.best_model_path) > 0
        ):
            self.best_model_path = self.trainer.checkpoint_callback.best_model_path
            log.info("Loading best epoch weights...")

    def test(self) -> None:
        """Test the model."""
        if not self.config.trainer.get("fast_dev_run"):
            log.info("Starting testing!")
            self.trainer.test(datamodule=self.datamodule, model=self.module, ckpt_path=self.best_model_path)

    def export(self) -> None:
        """Generate deployment models for the task."""
        if self.datamodule.class_to_idx is None:
            log.warning(
                "No `class_to_idx` found in the datamodule, class information will not be saved in the model.json"
            )
            idx_to_class = {}
        else:
            idx_to_class = {v: k for k, v in self.datamodule.class_to_idx.items()}

        # Get best model!
        if self.best_model_path is not None:
            log.info("Saving deployment model for %s checkpoint", self.best_model_path)

            module = self.module.__class__.load_from_checkpoint(
                self.best_model_path,
                model=self.module.model,
                optimizer=self.optimizer,
                lr_scheduler=self.scheduler,
                criterion=self.module.criterion,
                gradcam=False,
            )
        else:
            log.warning("No checkpoint callback found in the trainer, exporting the last model weights")
            module = self.module

        input_shapes = self.config.export.input_shapes

        # TODO: What happens if we have 64 precision?
        half_precision = "16" in self.trainer.precision

        example_input: torch.Tensor | None = None

        if hasattr(self.trainer, "datamodule") and hasattr(self.trainer.datamodule, "val_dataset"):
            # Retrieve a better input to evaluate fp16 performance or efficientnetb0 does not sometimes export properly
            example_input = self.trainer.datamodule.val_dataset[0][0]

        # Selected rtol and atol are quite high, this is mostly done for efficientnetb0 that seems to be
        # quite unstable in fp16
        self.model_json, export_paths = export_model(
            config=self.config,
            model=module.model,
            export_folder=self.export_folder,
            half_precision=half_precision,
            input_shapes=input_shapes,
            idx_to_class=idx_to_class,
            example_inputs=example_input,
            rtol=0.05,
            atol=0.01,
        )

        if len(export_paths) == 0:
            return

        with open(os.path.join(self.export_folder, self.deploy_info_file), "w") as f:
            json.dump(self.model_json, f)

    def generate_report(self) -> None:
        """Generate a report for the task."""
        if self.datamodule.class_to_idx is None:
            log.warning("No `class_to_idx` found in the datamodule, report will not be generated")
            return

        if isinstance(self.datamodule, MultilabelClassificationDataModule):
            log.warning("Report generation is not supported for multilabel classification tasks at the moment.")
            return

        log.info("Generating report!")
        if not self.run_test or self.config.trainer.get("fast_dev_run"):
            self.datamodule.setup(stage="test")

        # Deepcopy to remove the inference mode from gradients causing issues when loading checkpoints
        # TODO: Why deepcopy of module model removes ModelSignatureWrapper?
        self.module.model.instance = deepcopy(self.module.model.instance)
        if "16" in self.trainer.precision:
            log.warning("Gradcam is currently not supported with half precision, it will be disabled")
            self.module.gradcam = False
            self.gradcam = False

        predictions_outputs = self.trainer.predict(
            model=self.module, datamodule=self.datamodule, ckpt_path=self.best_model_path
        )
        if not predictions_outputs:
            log.warning("There is no prediction to generate the report. Skipping report generation.")
            return
        all_outputs = [x[0] for x in predictions_outputs]
        all_probs = [x[2] for x in predictions_outputs]
        if not all_outputs or not all_probs:
            log.warning("There is no prediction to generate the report. Skipping report generation.")
            return
        all_outputs = [item for sublist in all_outputs for item in sublist]
        all_probs = [item for sublist in all_probs for item in sublist]
        all_targets = [target.tolist() for im, target in self.datamodule.test_dataloader()]
        all_targets = [item for sublist in all_targets for item in sublist]

        if self.module.gradcam:
            grayscale_cams = [x[1] for x in predictions_outputs]
            grayscale_cams = [item for sublist in grayscale_cams for item in sublist]
            grayscale_cams = np.stack(grayscale_cams)  # N x H x W
        else:
            grayscale_cams = None

        # creating confusion matrix
        idx_to_class = {v: k for k, v in self.datamodule.class_to_idx.items()}
        _, self.report_confmat, accuracy = get_results(
            test_labels=all_targets,
            pred_labels=all_outputs,
            idx_to_labels=idx_to_class,
        )
        output_folder_test = "test"
        test_dataloader = self.datamodule.test_dataloader()
        test_dataset = cast(ImageClassificationListDataset, test_dataloader.dataset)
        self.res = pd.DataFrame(
            {
                "sample": list(test_dataset.x),
                "real_label": all_targets,
                "pred_label": all_outputs,
                "probability": all_probs,
            }
        )
        os.makedirs(output_folder_test, exist_ok=True)
        save_classification_result(
            results=self.res,
            output_folder=output_folder_test,
            confmat=self.report_confmat,
            accuracy=accuracy,
            test_dataloader=self.datamodule.test_dataloader(),
            config=self.config,
            output=self.output,
            grayscale_cams=grayscale_cams,
        )

        if len(self.logger) > 0:
            mflow_logger = get_mlflow_logger(trainer=self.trainer)
            tensorboard_logger = utils.get_tensorboard_logger(trainer=self.trainer)
            artifacts = glob.glob(os.path.join(output_folder_test, "**/*"), recursive=True)
            if self.config.core.get("upload_artifacts") and len(artifacts) > 0:
                if mflow_logger is not None:
                    log.info("Uploading artifacts to MLFlow")
                    for a in artifacts:
                        if os.path.isdir(a):
                            continue

                        dirname = Path(a).parent.name
                        mflow_logger.experiment.log_artifact(
                            run_id=mflow_logger.run_id,
                            local_path=a,
                            artifact_path=os.path.join("classification_output", dirname),
                        )
                if tensorboard_logger is not None:
                    log.info("Uploading artifacts to Tensorboard")
                    for a in artifacts:
                        if os.path.isdir(a):
                            continue

                        ext = os.path.splitext(a)[1].lower()

                        if ext in [".png", ".jpg", ".jpeg", ".bmp", ".tiff", ".tif", ".gif"]:
                            try:
                                img = cv2.imread(a)
                                img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
                            except cv2.error:
                                log.info("Could not upload artifact image %s", a)
                                continue
                            output_path = os.path.sep.join(a.split(os.path.sep)[-2:])
                            tensorboard_logger.experiment.add_image(output_path, img, 0, dataformats="HWC")
                        else:
                            utils.upload_file_tensorboard(a, tensorboard_logger)

    def freeze_layers_by_name(self, freeze_parameters_name: list[str]):
        """Freeze layers specified in freeze_parameters_name.

        Args:
            freeze_parameters_name: Layers that will be frozen during training.

        """
        count_frozen = 0
        for name, param in self.model.named_parameters():
            if any(x in name.split(".")[1] for x in freeze_parameters_name):
                log.debug("Freezing layer %s", name)
                param.requires_grad = False

            if not param.requires_grad:
                count_frozen += 1

        log.info("Frozen %d parameters", count_frozen)

    def freeze_parameters_by_index(self, freeze_parameters_index: list[int]):
        """Freeze parameters specified in freeze_parameters_name.

        Args:
            freeze_parameters_index: Indices of parameters that will be frozen during training.

        """
        if getattr(self.config.backbone, "freeze_parameters_name", None) is not None:
            log.warning(
                "Please be aware that some of the model's parameters have already been frozen using \
                the specified freeze_parameters_name. You are combining these two actions."
            )
        count_frozen = 0
        for i, (name, param) in enumerate(self.model.named_parameters()):
            if i in freeze_parameters_index:
                log.debug("Freezing layer %s", name)
                param.requires_grad = False

            if not param.requires_grad:
                count_frozen += 1

        log.info("Frozen %d parameters", count_frozen)


class SklearnClassification(Generic[SklearnClassificationDataModuleT], Task[SklearnClassificationDataModuleT]):
    """Sklearn classification task.

    Args:
        config: The experiment configuration
        device: The device to use. Defaults to None.
        output: Dictionary defining which kind of outputs to generate. Defaults to None.
        automatic_batch_size: Whether to automatically find the largest batch size that fits in memory.
        save_model_summary: Whether to save a model_summary.txt file containing the model summary.
        half_precision: Whether to use half precision during training.
        gradcam: Whether to compute gradcams for test results.
    """

    def __init__(
        self,
        config: DictConfig,
        output: DictConfig,
        device: str,
        automatic_batch_size: DictConfig,
        save_model_summary: bool = False,
        half_precision: bool = False,
        gradcam: bool = False,
    ):
        super().__init__(config=config)

        self._device = device
        self.output = output
        self._backbone: ModelSignatureWrapper
        self._trainer: SklearnClassificationTrainer
        self._model: ClassifierMixin
        self.metadata: dict[str, Any] = {
            "test_confusion_matrix": [],
            "test_accuracy": [],
            "test_results": [],
            "test_labels": [],
            "cams": [],
        }
        self.export_folder = "deployment_model"
        self.deploy_info_file = "model.json"
        self.train_dataloader_list: list[torch.utils.data.DataLoader] = []
        self.test_dataloader_list: list[torch.utils.data.DataLoader] = []
        self.automatic_batch_size = automatic_batch_size
        self.save_model_summary = save_model_summary
        self.half_precision = half_precision
        self.gradcam = gradcam

    @property
    def device(self) -> str:
        return self._device

    def prepare(self) -> None:
        """Prepare the experiment."""
        self.datamodule = self.config.datamodule

        self.backbone = self.config.backbone

        self.model = self.config.model

        # prepare_data() must be explicitly called if the task does not include a lightining training
        self.datamodule.prepare_data()
        self.datamodule.setup(stage="fit")

        self.trainer = self.config.trainer

    @property
    def model(self) -> ClassifierMixin:
        """sklearn.base.ClassifierMixin: The model."""
        return self._model

    @model.setter
    def model(self, model_config: DictConfig):
        """sklearn.base.ClassifierMixin: The model."""
        log.info("Instantiating model <%s>", model_config["_target_"])
        self._model = hydra.utils.instantiate(model_config)

    @property
    def backbone(self) -> ModelSignatureWrapper:
        """Backbone: The backbone."""
        return self._backbone

    @backbone.setter
    def backbone(self, backbone_config):
        """Load backbone."""
        if backbone_config.metadata.get("checkpoint"):
            log.info("Loading backbone from <%s>", backbone_config.metadata.checkpoint)
            self._backbone = torch.load(backbone_config.metadata.checkpoint)
        else:
            log.info("Loading backbone from <%s>", backbone_config.model["_target_"])
            self._backbone = hydra.utils.instantiate(backbone_config.model)

        self._backbone = ModelSignatureWrapper(self._backbone)
        self._backbone.eval()
        if self.half_precision:
            if self.device == "cpu":
                raise ValueError("Half precision is not supported on CPU")
            self._backbone.half()

            if self.gradcam:
                log.warning("Gradcam is currently not supported with half precision, it will be disabled")
                self.gradcam = False
        self._backbone.to(self.device)

    @property
    def trainer(self) -> SklearnClassificationTrainer:
        """Trainer: The trainer."""
        return self._trainer

    @trainer.setter
    def trainer(self, trainer_config: DictConfig) -> None:
        """Trainer: The trainer."""
        log.info("Instantiating trainer <%s>", trainer_config["_target_"])
        trainer = hydra.utils.instantiate(trainer_config, backbone=self.backbone, classifier=self.model)
        self._trainer = trainer

    @typing.no_type_check
    @automatic_datamodule_batch_size(batch_size_attribute_name="batch_size")
    def train(self) -> None:
        """Train the model."""
        log.info("Starting training...!")
        all_features = None
        all_labels = None

        class_to_keep = None

        self.train_dataloader_list = list(self.datamodule.train_dataloader())
        self.test_dataloader_list = list(self.datamodule.val_dataloader())

        if hasattr(self.datamodule, "class_to_keep_training") and self.datamodule.class_to_keep_training is not None:
            class_to_keep = self.datamodule.class_to_keep_training

        if self.save_model_summary:
            self.extract_model_summary(feature_extractor=self.backbone, dl=self.datamodule.full_dataloader())

        if hasattr(self.datamodule, "cache") and self.datamodule.cache:
            if self.config.trainer.iteration_over_training != 1:
                raise AttributeError("Cache is only supported when iteration over training is set to 1")

            full_dataloader = self.datamodule.full_dataloader()
            all_features, all_labels, _ = get_feature(
                feature_extractor=self.backbone, dl=full_dataloader, iteration_over_training=1
            )

            sorted_indices = np.argsort(full_dataloader.dataset.x)
            all_features = all_features[sorted_indices]
            all_labels = all_labels[sorted_indices]

        # cycle over all train/test split
        for train_dataloader, test_dataloader in zip(
            self.train_dataloader_list, self.test_dataloader_list, strict=False
        ):
            # Reinit classifier
            self.model = self.config.model
            self.trainer.change_classifier(self.model)

            # Train on current training set
            if all_features is not None and all_labels is not None:
                # Find which are the indices used to pass from the sorted list of string to the disordered one
                sorted_indices = np.argsort(np.concatenate([train_dataloader.dataset.x, test_dataloader.dataset.x]))
                revese_sorted_indices = np.argsort(sorted_indices)

                # Use these indices to correctly match the extracted features with the new file order
                all_features_sorted = all_features[revese_sorted_indices]
                all_labels_sorted = all_labels[revese_sorted_indices]

                train_len = len(train_dataloader.dataset.x)

                self.trainer.fit(
                    train_features=all_features_sorted[0:train_len], train_labels=all_labels_sorted[0:train_len]
                )

                _, pd_cm, accuracy, res, cams = self.trainer.test(
                    test_dataloader=test_dataloader,
                    test_features=all_features_sorted[train_len:],
                    test_labels=all_labels_sorted[train_len:],
                    class_to_keep=class_to_keep,
                    idx_to_class=train_dataloader.dataset.idx_to_class,
                    predict_proba=True,
                    gradcam=self.gradcam,
                )
            else:
                self.trainer.fit(train_dataloader=train_dataloader)
                _, pd_cm, accuracy, res, cams = self.trainer.test(
                    test_dataloader=test_dataloader,
                    class_to_keep=class_to_keep,
                    idx_to_class=train_dataloader.dataset.idx_to_class,
                    predict_proba=True,
                    gradcam=self.gradcam,
                )

            # save results
            self.metadata["test_confusion_matrix"].append(pd_cm)
            self.metadata["test_accuracy"].append(accuracy)
            self.metadata["test_results"].append(res)
            self.metadata["test_labels"].append(
                [
                    train_dataloader.dataset.idx_to_class[i] if i != -1 else "N/A"
                    for i in res["real_label"].unique().tolist()
                ]
            )
            self.metadata["cams"].append(cams)

    def extract_model_summary(
        self, feature_extractor: torch.nn.Module | BaseEvaluationModel, dl: torch.utils.data.DataLoader
    ) -> None:
        """Given a dataloader and a PyTorch model, use torchinfo to extract a summary of the model and save it
        to a file.

        Args:
            dl: PyTorch dataloader
            feature_extractor: PyTorch backbone
        """
        if isinstance(feature_extractor, TorchEvaluationModel | TorchscriptEvaluationModel):
            # TODO: I'm not sure torchinfo supports torchscript models
            # If we are working with torch based evaluation models we need to extract the model
            feature_extractor = feature_extractor.model

        for b in tqdm(dl):
            x1, _ = b

            if hasattr(feature_extractor, "parameters"):
                # Move input to the correct device
                parameter = next(feature_extractor.parameters())
                x1 = x1.to(parameter.device).to(parameter.dtype)
                x1 = x1[0].unsqueeze(0)  # Remove batch dimension

                model_info = None

                try:
                    try:
                        # TODO: Do we want to print the summary to the console as well?
                        model_info = summary(feature_extractor, input_data=(x1), verbose=0)  # type: ignore[arg-type]
                    except Exception:
                        log.warning(
                            "Failed to retrieve model summary using input data information, retrieving only "
                            "parameters information"
                        )
                        model_info = summary(feature_extractor, verbose=0)  # type: ignore[arg-type]
                except Exception as e:
                    # If for some reason the summary fails we don't want to stop the training
                    log.warning("Failed to retrieve model summary: %s", e)

                if model_info is not None:
                    with open("model_summary.txt", "w") as f:
                        f.write(str(model_info))
            else:
                log.warning("Failed to retrieve model summary, current model has no parameters")

            break

    @automatic_datamodule_batch_size(batch_size_attribute_name="batch_size")
    def train_full_data(self):
        """Train the model on train + validation."""
        # Reinit classifier
        self.model = self.config.model
        self.trainer.change_classifier(self.model)

        self.trainer.fit(train_dataloader=self.datamodule.full_dataloader())

    def test(self) -> None:
        """Skip test phase."""
        # we don't need test phase since sklearn trainer is already running test inside
        # train module to handle cross validation

    @typing.no_type_check
    @automatic_datamodule_batch_size(batch_size_attribute_name="batch_size")
    def test_full_data(self) -> None:
        """Test model trained on full dataset."""
        self.config.datamodule.class_to_idx = self.datamodule.full_dataset.class_to_idx
        self.config.datamodule.phase = "test"
        idx_to_class = self.datamodule.full_dataset.idx_to_class
        self.datamodule.setup("test")
        test_dataloader = self.datamodule.test_dataloader()

        if len(self.datamodule.data["samples"]) == 0:
            log.info("No test data, skipping test")
            return

        # Put backbone on the correct device as it may be moved after export
        self.backbone.to(self.device)
        _, pd_cm, accuracy, res, cams = self.trainer.test(
            test_dataloader=test_dataloader, idx_to_class=idx_to_class, predict_proba=True, gradcam=self.gradcam
        )

        output_folder_test = "test"

        os.makedirs(output_folder_test, exist_ok=True)

        save_classification_result(
            results=res,
            output_folder=output_folder_test,
            confmat=pd_cm,
            accuracy=accuracy,
            test_dataloader=test_dataloader,
            config=self.config,
            output=self.output,
            grayscale_cams=cams,
        )

    def export(self) -> None:
        """Generate deployment model for the task."""
        if self.config.export is None or len(self.config.export.types) == 0:
            log.info("No export type specified skipping export")
            return

        input_shapes = self.config.export.input_shapes

        idx_to_class = {v: k for k, v in self.datamodule.full_dataset.class_to_idx.items()}

        model_json, export_paths = export_model(
            config=self.config,
            model=self.backbone,
            export_folder=self.export_folder,
            half_precision=self.half_precision,
            input_shapes=input_shapes,
            idx_to_class=idx_to_class,
            pytorch_model_type="backbone",
        )

        dump(self.model, os.path.join(self.export_folder, "classifier.joblib"))

        if len(export_paths) > 0:
            with open(os.path.join(self.export_folder, self.deploy_info_file), "w") as f:
                json.dump(model_json, f)

    def generate_report(self) -> None:
        """Generate report for the task."""
        log.info("Generating report!")

        cm_list = []

        for count in range(len(self.metadata["test_accuracy"])):
            current_output_folder = f"{self.output.folder}_{count}"
            os.makedirs(current_output_folder, exist_ok=True)

            c_matrix = self.metadata["test_confusion_matrix"][count]
            cm_list.append(c_matrix)
            save_classification_result(
                results=self.metadata["test_results"][count],
                output_folder=current_output_folder,
                confmat=c_matrix,
                accuracy=self.metadata["test_accuracy"][count],
                test_dataloader=self.test_dataloader_list[count],
                config=self.config,
                output=self.output,
                grayscale_cams=self.metadata["cams"][count],
            )
        final_confusion_matrix = sum(cm_list)

        self.metadata["final_confusion_matrix"] = final_confusion_matrix
        # Save final conf matrix
        final_folder = f"{self.output.folder}"
        os.makedirs(final_folder, exist_ok=True)
        disp = ConfusionMatrixDisplay(
            confusion_matrix=np.array(final_confusion_matrix),
            display_labels=[x.replace("pred:", "") for x in final_confusion_matrix.columns.to_list()],
        )
        disp.plot(include_values=True, cmap=plt.cm.Greens, ax=None, colorbar=False, xticks_rotation=90)
        plt.title(f"Confusion Matrix (Accuracy: {(self.metadata['test_accuracy'][count] * 100):.2f}%)")
        plt.savefig(os.path.join(final_folder, "test_confusion_matrix.png"), bbox_inches="tight", pad_inches=0, dpi=300)
        plt.close()

    def execute(self) -> None:
        """Execute the experiment and all the steps."""
        self.prepare()
        self.train()
        if self.output.report:
            self.generate_report()
        self.train_full_data()
        if self.config.export is not None and len(self.config.export.types) > 0:
            self.export()
        if self.output.test_full_data:
            self.test_full_data()
        self.finalize()


class SklearnTestClassification(Evaluation[SklearnClassificationDataModuleT]):
    """Perform a test using an imported SklearnClassification pytorch model.

    Args:
        config: The experiment configuration
        output: where to save results
        model_path: path to trained model generated from SklearnClassification task.
        device: the device where to run the model (cuda or cpu)
        gradcam: Whether to compute gradcams
        **kwargs: Additional arguments to pass to the task
    """

    def __init__(
        self,
        config: DictConfig,
        output: DictConfig,
        model_path: str,
        device: str,
        gradcam: bool = False,
        **kwargs: Any,
    ):
        super().__init__(config=config, model_path=model_path, device=device, **kwargs)
        self.gradcam = gradcam
        self.output = output
        self._backbone: BaseEvaluationModel
        self._classifier: ClassifierMixin
        self.class_to_idx: dict[str, int]
        self.idx_to_class: dict[int, str]
        self.test_dataloader: torch.utils.data.DataLoader
        self.metadata: dict[str, Any] = {
            "test_confusion_matrix": None,
            "test_accuracy": None,
            "test_results": None,
            "test_labels": None,
            "cams": None,
        }

    def prepare(self) -> None:
        """Prepare the experiment."""
        super().prepare()

        idx_to_class = {}
        class_to_idx = {}
        for k, v in self.model_data["classes"].items():
            idx_to_class[int(k)] = v
            class_to_idx[v] = int(k)

        self.idx_to_class = idx_to_class
        self.class_to_idx = class_to_idx

        self.config.datamodule.class_to_idx = class_to_idx

        self.datamodule = self.config.datamodule
        # prepare_data() must be explicitly called because there is no lightning training
        self.datamodule.prepare_data()
        self.datamodule.setup(stage="test")

        # Configure trainer
        self.trainer = self.config.trainer

    @property
    def deployment_model(self):
        """Deployment model."""
        return None

    @deployment_model.setter
    def deployment_model(self, model_path: str):
        """Set backbone and classifier."""
        self.backbone = model_path  # type: ignore[assignment]
        # Load classifier
        self.classifier = os.path.join(Path(model_path).parent, "classifier.joblib")

    @property
    def classifier(self) -> ClassifierMixin:
        """Classifier: The classifier."""
        return self._classifier

    @classifier.setter
    def classifier(self, classifier_path: str) -> None:
        """Load classifier."""
        self._classifier = load(classifier_path)

    @property
    def backbone(self) -> BaseEvaluationModel:
        """Backbone: The backbone."""
        return self._backbone

    @backbone.setter
    def backbone(self, model_path: str) -> None:
        """Load backbone."""
        file_extension = os.path.splitext(model_path)[1]

        model_architecture = None
        if file_extension == ".pth":
            backbone_config_path = os.path.join(Path(model_path).parent, "model_config.yaml")
            log.info("Loading backbone from config")
            backbone_config = OmegaConf.load(backbone_config_path)

            if backbone_config.metadata.get("checkpoint"):
                log.info("Loading backbone from <%s>", backbone_config.metadata.checkpoint)
                model_architecture = torch.load(backbone_config.metadata.checkpoint)
            else:
                log.info("Loading backbone from <%s>", backbone_config.model["_target_"])
                model_architecture = hydra.utils.instantiate(backbone_config.model)

        self._backbone = import_deployment_model(
            model_path=model_path,
            device=self.device,
            inference_config=self.config.inference,
            model_architecture=model_architecture,
        )

        if self.gradcam and not isinstance(self._backbone, TorchEvaluationModel):
            log.warning("Gradcam is supported only for pytorch models. Skipping gradcam")
            self.gradcam = False

    @property
    def trainer(self) -> SklearnClassificationTrainer:
        """Trainer: The trainer."""
        return self._trainer

    @trainer.setter
    def trainer(self, trainer_config: DictConfig) -> None:
        """Trainer: The trainer."""
        log.info("Instantiating trainer <%s>", trainer_config["_target_"])

        if self.backbone.training:
            self.backbone.eval()

        trainer = hydra.utils.instantiate(trainer_config, backbone=self.backbone, classifier=self.classifier)
        self._trainer = trainer

    @automatic_datamodule_batch_size(batch_size_attribute_name="batch_size")
    def test(self) -> None:
        """Run the test."""
        self.test_dataloader = self.datamodule.test_dataloader()

        _, pd_cm, accuracy, res, cams = self.trainer.test(
            test_dataloader=self.test_dataloader,
            idx_to_class=self.idx_to_class,
            predict_proba=True,
            gradcam=self.gradcam,
        )

        # save results
        self.metadata["test_confusion_matrix"] = pd_cm
        self.metadata["test_accuracy"] = accuracy
        self.metadata["test_results"] = res
        self.metadata["test_labels"] = [
            self.idx_to_class[i] if i != -1 else "N/A" for i in res["real_label"].unique().tolist()
        ]
        self.metadata["cams"] = cams

    def generate_report(self) -> None:
        """Generate a report for the task."""
        log.info("Generating report!")
        os.makedirs(self.output.folder, exist_ok=True)
        save_classification_result(
            results=self.metadata["test_results"],
            output_folder=self.output.folder,
            confmat=self.metadata["test_confusion_matrix"],
            accuracy=self.metadata["test_accuracy"],
            test_dataloader=self.test_dataloader,
            config=self.config,
            output=self.output,
            grayscale_cams=self.metadata["cams"],
        )

    def execute(self) -> None:
        """Execute the experiment and all the steps."""
        self.prepare()
        self.test()
        if self.output.report:
            self.generate_report()
        self.finalize()


class ClassificationEvaluation(Evaluation[ClassificationDataModuleT]):
    """Perform a test on an imported Classification pytorch model.

    Args:
        config: Task configuration
        output: Configuration for the output
        model_path: Path to pytorch .pt model file
        report: Whether to generate the report of the predictions
        gradcam: Whether to compute gradcams
        device: Device to use for evaluation. If None, the device is automatically determined

    """

    def __init__(
        self,
        config: DictConfig,
        output: DictConfig,
        model_path: str,
        report: bool = True,
        gradcam: bool = False,
        device: str | None = None,
    ):
        super().__init__(config=config, model_path=model_path, device=device)
        self.report_path = "test_output"
        self.output = output
        self.report = report
        self.gradcam = gradcam
        self.cam: GradCAM

    def get_torch_model(self, model_config: DictConfig) -> nn.Module:
        """Instantiate the torch model from the config."""
        pre_classifier = self.get_pre_classifier(model_config)
        classifier = self.get_classifier(model_config)
        log.info("Instantiating backbone <%s>", model_config.model["_target_"])

        return hydra.utils.instantiate(
            model_config.model, classifier=classifier, pre_classifier=pre_classifier, _convert_="partial"
        )

    def get_pre_classifier(self, model_config: DictConfig) -> nn.Module:
        """Instantiate the pre-classifier from the config."""
        if "pre_classifier" in model_config and model_config.pre_classifier is not None:
            log.info("Instantiating pre_classifier <%s>", model_config.pre_classifier["_target_"])
            pre_classifier = hydra.utils.instantiate(model_config.pre_classifier, _convert_="partial")
        else:
            log.info("No pre-classifier found in config: instantiate a torch.nn.Identity instead")
            pre_classifier = nn.Identity()

        return pre_classifier

    def get_classifier(self, model_config: DictConfig) -> nn.Module:
        """Instantiate the classifier from the config."""
        if "classifier" in model_config:
            log.info("Instantiating classifier <%s>", model_config.classifier["_target_"])
            return hydra.utils.instantiate(
                model_config.classifier, out_features=len(self.model_data["classes"]), _convert_="partial"
            )

        raise ValueError("A `classifier` definition must be specified in the config")

    @property
    def deployment_model(self) -> BaseEvaluationModel:
        """Deployment model."""
        return self._deployment_model

    @deployment_model.setter
    def deployment_model(self, model_path: str):
        """Set the deployment model."""
        file_extension = os.path.splitext(model_path)[1]
        model_architecture = None
        if file_extension == ".pth":
            model_config = OmegaConf.load(os.path.join(Path(model_path).parent, "model_config.yaml"))

            if not isinstance(model_config, DictConfig):
                raise ValueError(f"The model config must be a DictConfig, got {type(model_config)}")

            model_architecture = self.get_torch_model(model_config)

        self._deployment_model = import_deployment_model(
            model_path=model_path,
            device=self.device,
            inference_config=self.config.inference,
            model_architecture=model_architecture,
        )

        if self.gradcam and not isinstance(self.deployment_model, TorchEvaluationModel):
            log.warning("To compute gradcams you need to provide the path to an exported .pth state_dict file")
            self.gradcam = False

    def prepare(self) -> None:
        """Prepare the evaluation."""
        super().prepare()
        self.datamodule = self.config.datamodule
        self.datamodule.class_to_idx = {v: int(k) for k, v in self.model_data["classes"].items()}
        self.datamodule.num_classes = len(self.datamodule.class_to_idx)

        # prepare_data() must be explicitly called because there is no training
        self.datamodule.prepare_data()
        self.datamodule.setup(stage="test")

    def prepare_gradcam(self) -> None:
        """Initializing gradcam for the predictions."""
        if not hasattr(self.deployment_model.model, "features_extractor"):
            log.warning("Gradcam not implemented for this backbone, it will not be computed")
            self.gradcam = False
            return

        if isinstance(self.deployment_model.model.features_extractor, timm.models.resnet.ResNet):
            target_layers = [cast(BaseNetworkBuilder, self.deployment_model.model).features_extractor.layer4[-1]]  # type: ignore[index]
            self.cam = GradCAM(
                model=self.deployment_model.model,
                target_layers=target_layers,
            )
            for p in self.deployment_model.model.features_extractor.layer4[-1].parameters():
                p.requires_grad = True
        elif is_vision_transformer(cast(BaseNetworkBuilder, self.deployment_model.model).features_extractor):
            self.grad_rollout = VitAttentionGradRollout(cast(nn.Module, self.deployment_model.model))
        else:
            log.warning("Gradcam not implemented for this backbone, it will not be computed")
            self.gradcam = False

    @automatic_datamodule_batch_size(batch_size_attribute_name="batch_size")
    def test(self) -> None:
        """Perform test."""
        log.info("Running test")
        test_dataloader = self.datamodule.test_dataloader()

        image_labels = []
        probabilities = []
        predicted_classes = []
        grayscale_cams_list = []

        if self.gradcam:
            self.prepare_gradcam()

        with torch.set_grad_enabled(self.gradcam):
            for batch_item in tqdm(test_dataloader):
                im, target = batch_item
                im = im.to(device=self.device, dtype=self.deployment_model.model_dtype).detach()

                if self.gradcam:
                    # When gradcam is used we need to remove gradients
                    outputs = self.deployment_model(im).detach()
                else:
                    outputs = self.deployment_model(im)

                probs = torch.softmax(outputs, dim=1)
                preds = torch.max(probs, dim=1).indices

                probabilities.append(probs.tolist())
                predicted_classes.append(preds.tolist())
                image_labels.extend(target.tolist())
                if self.gradcam and hasattr(self.deployment_model.model, "features_extractor"):
                    with torch.inference_mode(False):
                        im = im.clone()
                        if isinstance(self.deployment_model.model.features_extractor, timm.models.resnet.ResNet):
                            grayscale_cam = self.cam(input_tensor=im, targets=None)
                            grayscale_cams_list.append(torch.from_numpy(grayscale_cam))
                        elif is_vision_transformer(
                            cast(BaseNetworkBuilder, self.deployment_model.model).features_extractor
                        ):
                            grayscale_cam_low_res = self.grad_rollout(input_tensor=im, targets_list=preds.tolist())
                            orig_shape = grayscale_cam_low_res.shape
                            new_shape = (orig_shape[0], im.shape[2], im.shape[3])
                            zoom_factors = tuple(np.array(new_shape) / np.array(orig_shape))
                            grayscale_cam = ndimage.zoom(grayscale_cam_low_res, zoom_factors, order=1)
                            grayscale_cams_list.append(torch.from_numpy(grayscale_cam))

        grayscale_cams: torch.Tensor | None = None
        if self.gradcam:
            grayscale_cams = torch.cat(grayscale_cams_list, dim=0)

        predicted_classes = [item for sublist in predicted_classes for item in sublist]
        probabilities = [max(item) for sublist in probabilities for item in sublist]
        if self.datamodule.class_to_idx is not None:
            idx_to_class = {v: k for k, v in self.datamodule.class_to_idx.items()}
        else:
            idx_to_class = None

        _, pd_cm, test_accuracy = get_results(
            test_labels=image_labels,
            pred_labels=predicted_classes,
            idx_to_labels=idx_to_class,
        )

        res = pd.DataFrame(
            {
                "sample": list(test_dataloader.dataset.x),  # type: ignore[attr-defined]
                "real_label": image_labels,
                "pred_label": predicted_classes,
                "probability": probabilities,
            }
        )

        log.info("Avg classification accuracy: %s", test_accuracy)

        self.res = pd.DataFrame(
            {
                "sample": list(test_dataloader.dataset.x),  # type: ignore[attr-defined]
                "real_label": image_labels,
                "pred_label": predicted_classes,
                "probability": probabilities,
            }
        )

        # save results
        self.metadata["test_confusion_matrix"] = pd_cm
        self.metadata["test_accuracy"] = test_accuracy
        self.metadata["predictions"] = predicted_classes
        self.metadata["test_results"] = res
        self.metadata["probabilities"] = probabilities
        self.metadata["test_labels"] = image_labels
        self.metadata["grayscale_cams"] = grayscale_cams

    def generate_report(self) -> None:
        """Generate a report for the task."""
        log.info("Generating report!")
        os.makedirs(self.report_path, exist_ok=True)

        save_classification_result(
            results=self.metadata["test_results"],
            output_folder=self.report_path,
            confmat=self.metadata["test_confusion_matrix"],
            accuracy=self.metadata["test_accuracy"],
            test_dataloader=self.datamodule.test_dataloader(),
            config=self.config,
            output=self.output,
            grayscale_cams=self.metadata["grayscale_cams"],
        )

    def execute(self) -> None:
        """Execute the evaluation."""
        self.prepare()
        self.test()
        if self.report:
            self.generate_report()
        self.finalize()



from __future__ import annotations

from typing import Any, cast

import numpy as np
import timm
import torch
import torchmetrics
import torchmetrics.functional as TMF
from pytorch_grad_cam import GradCAM
from scipy import ndimage
from torch import nn, optim

from quadra.models.classification import BaseNetworkBuilder
from quadra.modules.base import BaseLightningModule
from quadra.utils.models import is_vision_transformer
from quadra.utils.utils import get_logger
from quadra.utils.vit_explainability import VitAttentionGradRollout

log = get_logger(__name__)


class ClassificationModule(BaseLightningModule):
    """Lightning module for classification tasks.

    Args:
        model: Feature extractor as PyTorch `torch.nn.Module`
        criterion: the loss to be applied as a PyTorch `torch.nn.Module`.
        optimizer: optimizer of the training. Defaults to None.
        lr_scheduler: Pytorch learning rate scheduler.
            If None a default ReduceLROnPlateau is used.
            Defaults to None.
        lr_scheduler_interval: the learning rate scheduler interval.
            Defaults to "epoch".
        gradcam (bool): Whether to compute gradcam during prediction step
    """

    def __init__(
        self,
        model: nn.Module,
        criterion: nn.Module,
        optimizer: None | optim.Optimizer = None,
        lr_scheduler: None | object = None,
        lr_scheduler_interval: str | None = "epoch",
        gradcam: bool = False,
    ):
        super().__init__(model, optimizer, lr_scheduler, lr_scheduler_interval)

        self.criterion = criterion
        self.gradcam = gradcam
        self.train_acc = torchmetrics.Accuracy()
        self.val_acc = torchmetrics.Accuracy()
        self.test_acc = torchmetrics.Accuracy()
        self.cam: GradCAM | None = None
        self.grad_rollout: VitAttentionGradRollout | None = None

        if not isinstance(self.model.features_extractor, timm.models.resnet.ResNet) and not is_vision_transformer(
            cast(BaseNetworkBuilder, self.model).features_extractor
        ):
            log.warning(
                "Backbone not compatible with gradcam. Only timm ResNets, timm ViTs and TorchHub dinoViTs supported",
            )
            self.gradcam = False

        self.original_requires_grads: list[bool] = []

    def forward(self, x: torch.Tensor):
        return self.model(x)

    def training_step(self, batch: tuple[torch.Tensor, torch.Tensor], batch_idx: int):
        # pylint: disable=unused-argument
        im, target = batch
        outputs = self(im)
        loss = self.criterion(outputs, target)

        self.log_dict(
            {"train_loss": loss},
            on_epoch=True,
            on_step=True,
            logger=True,
            prog_bar=True,
        )
        self.log_dict(
            {"train_acc": self.train_acc(outputs.argmax(1), target)},
            on_step=False,
            on_epoch=True,
            logger=True,
            prog_bar=True,
        )
        return loss

    def validation_step(self, batch: tuple[torch.Tensor, torch.Tensor], batch_idx: int):
        # pylint: disable=unused-argument
        im, target = batch
        outputs = self(im)
        loss = self.criterion(outputs, target)

        self.log_dict(
            {"val_loss": loss},
            on_epoch=True,
            on_step=True,
            logger=True,
            prog_bar=True,
        )
        self.log_dict(
            {"val_acc": self.val_acc(outputs.argmax(1), target)},
            on_step=False,
            on_epoch=True,
            logger=True,
            prog_bar=True,
        )
        return loss

    def test_step(self, batch: tuple[torch.Tensor, torch.Tensor], batch_idx: int):
        # pylint: disable=unused-argument
        im, target = batch
        outputs = self(im)

        loss = self.criterion(outputs, target)

        self.log_dict(
            {"test_loss": loss},
            on_epoch=True,
            on_step=True,
            logger=True,
            prog_bar=False,
        )
        self.log_dict(
            {"test_acc": self.test_acc(outputs.argmax(1), target)},
            on_step=False,
            on_epoch=True,
            logger=True,
            prog_bar=False,
        )

    def prepare_gradcam(self) -> None:
        """Instantiate gradcam handlers."""
        if isinstance(self.model.features_extractor, timm.models.resnet.ResNet):
            target_layers = [cast(BaseNetworkBuilder, self.model).features_extractor.layer4[-1]]

            self.cam = GradCAM(
                model=self.model,
                target_layers=target_layers,
            )
            # Activating gradients
            for p in self.model.features_extractor.layer4[-1].parameters():
                p.requires_grad = True
        elif is_vision_transformer(cast(BaseNetworkBuilder, self.model).features_extractor):
            self.grad_rollout = VitAttentionGradRollout(self.model)
        else:
            log.warning("Gradcam not implemented for this backbone, it won't be computed")
            self.original_requires_grads.clear()
            self.gradcam = False

    def on_predict_start(self) -> None:
        """If gradcam, prepares gradcam and saves params requires_grad state."""
        if self.gradcam:
            # Saving params requires_grad state
            for p in self.model.parameters():
                self.original_requires_grads.append(p.requires_grad)
            self.prepare_gradcam()

        return super().on_predict_start()

    def on_predict_end(self) -> None:
        """If we computed gradcam, requires_grad values are reset to original value."""
        if self.gradcam:
            # Get back to initial state
            for i, p in enumerate(self.model.parameters()):
                p.requires_grad = self.original_requires_grads[i]

            # We are using GradCAM package only for resnets at the moment
            if isinstance(self.model.features_extractor, timm.models.resnet.ResNet) and self.cam is not None:
                # Needed to solve jitting bug
                self.cam.activations_and_grads.release()
            elif (
                is_vision_transformer(cast(BaseNetworkBuilder, self.model).features_extractor)
                and self.grad_rollout is not None
            ):
                for handle in self.grad_rollout.f_hook_handles:
                    handle.remove()
                for handle in self.grad_rollout.b_hook_handles:
                    handle.remove()

        return super().on_predict_end()

    # pylint: disable=unused-argument
    def predict_step(self, batch: Any, batch_idx: int, dataloader_idx: int = 0) -> Any:
        """Prediction step.

        Args:
            batch: Tuple composed by (image, target)
            batch_idx: Batch index
            dataloader_idx: Dataloader index
        Returns:
            Tuple containing:
                predicted_classes: indexes of predicted classes
                grayscale_cam: gray scale gradcams
        """
        im, _ = batch
        outputs = self(im)
        probs = torch.softmax(outputs, dim=1)
        predicted_classes = torch.max(probs, dim=1).indices.tolist()
        if self.gradcam:
            # inference_mode set to false because gradcam needs gradients
            with torch.inference_mode(False):
                im = im.clone()

                if isinstance(self.model.features_extractor, timm.models.resnet.ResNet) and self.cam:
                    grayscale_cam = self.cam(input_tensor=im, targets=None)
                elif (
                    is_vision_transformer(cast(BaseNetworkBuilder, self.model).features_extractor) and self.grad_rollout
                ):
                    grayscale_cam_low_res = self.grad_rollout(input_tensor=im, targets_list=predicted_classes)
                    orig_shape = grayscale_cam_low_res.shape
                    new_shape = (orig_shape[0], im.shape[2], im.shape[3])
                    zoom_factors = tuple(np.array(new_shape) / np.array(orig_shape))
                    grayscale_cam = ndimage.zoom(grayscale_cam_low_res, zoom_factors, order=1)
        else:
            grayscale_cam = None
        return predicted_classes, grayscale_cam, torch.max(probs, dim=1)[0].tolist()


class MultilabelClassificationModule(BaseLightningModule):
    """SklearnClassification model: train a generic SklearnClassification model for a multilabel
    problem.

    Args:
        model: Feature extractor as PyTorch `torch.nn.Module`
        criterion: the loss to be applied as a PyTorch `torch.nn.Module`.
        optimizer: optimizer of the training. Defaults to None.
        lr_scheduler: Pytorch learning rate scheduler.
            If None a default ReduceLROnPlateau is used.
            Defaults to None.
        lr_scheduler_interval: the learning rate scheduler interval.
            Defaults to "epoch".
    """

    def __init__(
        self,
        model: nn.Sequential,
        criterion: nn.Module,
        optimizer: None | optim.Optimizer = None,
        lr_scheduler: None | object = None,
        lr_scheduler_interval: str | None = "epoch",
        gradcam: bool = False,
    ):
        super().__init__(model, optimizer, lr_scheduler, lr_scheduler_interval)
        self.criterion = criterion
        self.gradcam = gradcam

        # TODO: can we use gradcam with more backbones?
        if self.gradcam:
            if not isinstance(model[0].features_extractor, timm.models.resnet.ResNet):
                log.warning(
                    "Backbone must be compatible with gradcam, at the moment only ResNets supported, disabling gradcam"
                )
                self.gradcam = False
            else:
                target_layers = [model[0].features_extractor.layer4[-1]]
                self.cam = GradCAM(model=model, target_layers=target_layers)

    def forward(self, x):
        return self.model(x)

    def training_step(self, batch, batch_idx):
        # pylint: disable=unused-argument
        im, target = batch
        outputs = self(im)
        with torch.no_grad():
            outputs_sig = torch.sigmoid(outputs)
        loss = self.criterion(outputs, target)

        self.log_dict(
            {
                "t_loss": loss,
                "t_map": TMF.label_ranking_average_precision(outputs_sig, target.bool()),
                "t_f1": TMF.f1_score(outputs_sig, target.bool(), average="samples"),
            },
            on_epoch=True,
            on_step=False,
            logger=True,
            prog_bar=True,
        )
        return loss

    def validation_step(self, batch, batch_idx):
        # pylint: disable=unused-argument
        im, target = batch
        outputs = self(im)
        with torch.no_grad():
            outputs_sig = torch.sigmoid(outputs)
        loss = self.criterion(outputs, target)

        self.log_dict(
            {
                "val_loss": loss,
                "val_map": TMF.label_ranking_average_precision(outputs_sig, target.bool()),
                "val_f1": TMF.f1_score(outputs_sig, target.bool(), average="samples"),
            },
            on_epoch=True,
            on_step=False,
            logger=True,
            prog_bar=True,
        )
        return loss

    def test_step(self, batch, batch_idx):
        # pylint: disable=unused-argument
        im, target = batch
        outputs = self(im)
        with torch.no_grad():
            outputs_sig = torch.sigmoid(outputs)
        loss = self.criterion(outputs, target)

        self.log_dict(
            {
                "test_loss": loss,
                "test_map": TMF.label_ranking_average_precision(outputs_sig, target.bool()),
                "test_f1": TMF.f1_score(outputs_sig, target.bool(), average="samples"),
            },
            on_epoch=True,
            on_step=True,
            logger=True,
            prog_bar=False,
        )
        return loss



import pytest
import torch
from anomalib.models.draem.torch_model import DraemModel
from anomalib.models.efficient_ad.torch_model import EfficientAdModel
from anomalib.models.padim.torch_model import PadimModel
from anomalib.models.patchcore.torch_model import PatchcoreModel


@pytest.fixture
def padim_resnet18():
    """Yield a padim model with resnet18 encoder."""
    yield PadimModel(
        input_size=[224, 224],  # TODO: This is hardcoded may be not a good idea
        backbone="resnet18",
        layers=["layer1", "layer2", "layer3"],
        pretrained_weights=None,
        tied_covariance=False,
        pre_trained=False,
    )


@torch.inference_mode()
def _initialize_patchcore_model(patchcore_model: PatchcoreModel, coreset_sampling_ratio: float = 0.1) -> PatchcoreModel:
    """Initialize a Patchcore model by simulating a training step.

    Args:
        patchcore_model: Patchcore model to initialize
        coreset_sampling_ratio: Coreset sampling ratio to use for the initialization

    Returns:
        Patchcore model with initialized memory bank
    """
    with torch.no_grad():
        training_features = None
        random_input = torch.rand([1, 3, *patchcore_model.input_size])

        if training_features is None:
            training_features = patchcore_model(random_input)
        else:
            training_features = torch.cat([training_features, patchcore_model(random_input)], dim=0)

        patchcore_model.eval()
        patchcore_model.subsample_embedding(training_features, sampling_ratio=coreset_sampling_ratio)

        # Simulate a memory bank with 5 images, at the current stage patchcore onnx export is not handling
        # large memory banks well, so we are using a small one for the benchmark
        memory_bank_number, memory_bank_n_features = patchcore_model.memory_bank.shape
        patchcore_model.memory_bank = torch.rand([5 * memory_bank_number, memory_bank_n_features])
        patchcore_model.train()

    return patchcore_model


@pytest.fixture
def patchcore_resnet18():
    """Yield a patchcore model with resnet18 encoder."""
    model = PatchcoreModel(
        input_size=[224, 224],  # TODO: This is hardcoded may be not a good idea
        backbone="resnet18",
        layers=["layer2", "layer3"],
        pre_trained=False,
    )

    yield _initialize_patchcore_model(model)


@pytest.fixture
def draem():
    """Yield a draem model."""
    yield DraemModel()


@pytest.fixture
def efficient_ad_small():
    """Yield a draem model."""

    class EfficientAdForwardWrapper(EfficientAdModel):
        """Wrap the forward method to avoid passing optional parameters."""

        def forward(self, x):
            return super().forward(x, None)

    model = EfficientAdForwardWrapper(
        teacher_out_channels=384,
        input_size=[256, 256],  # TODO: This is hardcoded may be not a good idea
        pretrained_teacher_type="nelson",
    )

    yield model



from __future__ import annotations

import shutil
from dataclasses import dataclass
from pathlib import Path
from typing import Any

import cv2
import numpy as np
import pytest

from quadra.utils.tests.helpers import _random_image


@dataclass
class SegmentationDatasetArguments:
    """Segmentation dataset arguments.

    Args:
        train_samples: List of samples per class in train set, element at index 0 are good samples
        val_samples: List of samples per class in validation set, same as above.
        test_samples: List of samples per class in test set, same as above.
        classes: Optional list of class names, must be equal to len(train_samples) - 1
    """

    train_samples: list[int]
    val_samples: list[int] | None = None
    test_samples: list[int] | None = None
    classes: list[str] | None = None


def _build_segmentation_dataset(
    tmp_path: Path, dataset_arguments: SegmentationDatasetArguments
) -> tuple[str, SegmentationDatasetArguments, dict[str, int]]:
    """Generate segmentation dataset.

    Args:
        tmp_path: path to temporary directory
        dataset_arguments: dataset arguments

    Returns:
        Tuple containing path to dataset, dataset arguments and class to index mapping
    """
    train_samples = dataset_arguments.train_samples
    val_samples = dataset_arguments.val_samples
    test_samples = dataset_arguments.test_samples
    classes = (
        dataset_arguments.classes if dataset_arguments.classes else list(range(1, len(dataset_arguments.train_samples)))
    )

    segmentation_dataset_path = tmp_path / "segmentation_dataset"
    segmentation_dataset_path.mkdir()
    images_path = segmentation_dataset_path / "images"
    masks_path = segmentation_dataset_path / "masks"
    images_path.mkdir(parents=True)
    masks_path.mkdir(parents=True)
    class_to_idx = {class_name: i + 1 for i, class_name in enumerate(classes)}
    classes = [0] + classes

    counter = 0
    for split_name, split_samples in zip(["train", "val", "test"], [train_samples, val_samples, test_samples]):
        if split_samples is None:
            continue

        with open(segmentation_dataset_path / f"{split_name}.txt", "w") as split_file:
            for class_name, samples in zip(classes, split_samples):
                for _ in range(samples):
                    image = _random_image(size=(224, 224))
                    mask = np.zeros((224, 224), dtype=np.uint8)
                    if class_name != 0:
                        mask[100:150, 100:150] = class_to_idx[class_name]
                    image_path = images_path / f"{class_name}_{counter}.png"
                    mask_path = masks_path / f"{class_name}_{counter}.png"
                    cv2.imwrite(str(image_path), image)
                    cv2.imwrite(str(mask_path), mask)
                    split_file.write(f"images/{image_path.name}\n")
                    counter += 1

    return str(segmentation_dataset_path), dataset_arguments, class_to_idx


@pytest.fixture
def segmentation_dataset(
    tmp_path: Path, dataset_arguments: SegmentationDatasetArguments
) -> tuple[str, SegmentationDatasetArguments, dict[str, int]]:
    """Fixture to dinamically generate a segmentation dataset. By default generated images are 224x224 pixels
        and associated masks contains a 50x50 pixels square with the corresponding image class, so at the current stage
        is not possible to have images with multiple annotations. Split files are saved as train.txt,
        val.txt and test.txt.

    Args:
        tmp_path: path to temporary directory
        dataset_arguments: dataset arguments

    Yields:
        Tuple containing path to dataset, dataset arguments and class to index mapping
    """
    yield _build_segmentation_dataset(tmp_path, dataset_arguments)
    if tmp_path.exists():
        shutil.rmtree(tmp_path)


@pytest.fixture(
    params=[
        SegmentationDatasetArguments(
            **{"train_samples": [3, 2], "val_samples": [2, 2], "test_samples": [1, 1], "classes": ["bad"]}
        )
    ]
)
def base_binary_segmentation_dataset(
    tmp_path: Path, request: Any
) -> tuple[str, SegmentationDatasetArguments, dict[str, int]]:
    """Generate a base binary segmentation dataset with the following structure:
        - 3 good and 2 bad samples in train set
        - 2 good and 2 bad samples in validation set
        - 11 good and 1 bad sample in test set
        - 2 classes: good and bad.

    Args:
        tmp_path: path to temporary directory
        request: pytest request

    Yields:
        Tuple containing path to dataset, dataset arguments and class to index mapping
    """
    yield _build_segmentation_dataset(tmp_path, request.param)
    if tmp_path.exists():
        shutil.rmtree(tmp_path)


@pytest.fixture(
    params=[
        SegmentationDatasetArguments(
            **{
                "train_samples": [2, 2, 2],
                "val_samples": [2, 2, 2],
                "test_samples": [1, 1, 1],
                "classes": ["defect_1", "defect_2"],
            }
        )
    ]
)
def base_multiclass_segmentation_dataset(
    tmp_path: Path, request: Any
) -> tuple[str, SegmentationDatasetArguments, dict[str, int]]:
    """Generate a base binary segmentation dataset with the following structure:
        - 2 good, 2 defect_1 and 2 defect_2 samples in train set
        - 2 good, 2 defect_1 and 2 defect_2 samples in validation set
        - 1 good, 1 defect_1 and 1 defect_2 sample in test set
        - 3 classes: good, defect_1 and defect_2.

    Args:
        tmp_path: path to temporary directory
        request: pytest request

    Yields:
        Tuple containing path to dataset, dataset arguments and class to index mapping
    """
    yield _build_segmentation_dataset(tmp_path, request.param)
    if tmp_path.exists():
        shutil.rmtree(tmp_path)



from __future__ import annotations

import os
import shutil
from typing import Any

import pandas as pd
from sklearn.model_selection import train_test_split
from torchvision.datasets.utils import download_and_extract_archive

from quadra.datamodules import ClassificationDataModule, SSLDataModule
from quadra.utils.utils import get_logger

IMAGENETTE_LABEL_MAPPER = {
    "n01440764": "tench",
    "n02102040": "english_springer",
    "n02979186": "cassette_player",
    "n03000684": "chain_saw",
    "n03028079": "church",
    "n03394916": "french_horn",
    "n03417042": "garbage_truck",
    "n03425413": "gas_pump",
    "n03445777": "golf_ball",
    "n03888257": "parachute",
}

DEFAULT_CLASS_TO_IDX = {cl: idx for idx, cl in enumerate(sorted(IMAGENETTE_LABEL_MAPPER.values()))}

log = get_logger(__name__)


class ImagenetteClassificationDataModule(ClassificationDataModule):
    """Initializes the classification data module for Imagenette dataset.

    Args:
        data_path: Path to the dataset.
        name: Name of the dataset.
        imagenette_version: Version of the Imagenette dataset. Can be 320 or 160 or full.
        force_download: If True, the dataset will be downloaded even if the data_path already exists. The data_path
            will be deleted and recreated.
        class_to_idx: Dictionary mapping class names to class indices.
        **kwargs: Keyword arguments for the ClassificationDataModule.
    """

    def __init__(
        self,
        data_path: str,
        name: str = "imagenette_classification_datamodule",
        imagenette_version: str = "320",
        force_download: bool = False,
        class_to_idx: dict[str, int] | None = None,
        **kwargs: Any,
    ):
        if imagenette_version not in ["320", "160", "full"]:
            raise ValueError(f"imagenette_version must be one of 320, 160 or full. Got {imagenette_version} instead.")

        if imagenette_version == "full":
            imagenette_version = ""
        else:
            imagenette_version = f"-{imagenette_version}"

        self.download_url = f"https://s3.amazonaws.com/fast-ai-imageclas/imagenette2{imagenette_version}.tgz"
        self.force_download = force_download
        self.imagenette_version = imagenette_version

        if class_to_idx is None:
            class_to_idx = DEFAULT_CLASS_TO_IDX

        super().__init__(
            data_path=data_path,
            name=name,
            test_split_file=None,
            train_split_file=None,
            val_size=None,
            class_to_idx=class_to_idx,
            **kwargs,
        )

    def download_data(self, download_url: str, force_download: bool = False) -> None:
        """Download the Imagenette dataset.

        Args:
            download_url: Dataset download url.
            force_download: If True, the dataset will be downloaded even if the data_path already exists. The data_path
                will be removed.
        """
        if os.path.exists(self.data_path):
            if force_download:
                log.info("The path %s already exists. Removing it and downloading the dataset again.", self.data_path)
                shutil.rmtree(self.data_path)
            else:
                log.info("The path %s already exists. Skipping download.", self.data_path)
                return

        log.info("Downloading and extracting Imagenette dataset to %s", self.data_path)
        download_and_extract_archive(download_url, self.data_path, remove_finished=True)

    def _prepare_data(self) -> None:
        """Prepares the data for the data module."""
        self.download_data(download_url=self.download_url, force_download=self.force_download)
        self.data_path = os.path.join(self.data_path, f"imagenette2{self.imagenette_version}")

        train_images_and_targets, class_to_idx = self._find_images_and_targets(os.path.join(self.data_path, "train"))
        self.class_to_idx = {IMAGENETTE_LABEL_MAPPER[k]: v for k, v in class_to_idx.items()}

        samples_train, targets_train = [], []
        idx_to_class = {v: k for k, v in self.class_to_idx.items()}
        for image, target in train_images_and_targets:
            samples_train.append(image)
            targets_train.append(idx_to_class[target])

        samples_train, samples_val, targets_train, targets_val = train_test_split(
            samples_train,
            targets_train,
            test_size=self.val_size,
            random_state=self.seed,
            stratify=targets_train,
        )

        test_images_and_targets, _ = self._find_images_and_targets(os.path.join(self.data_path, "val"))
        samples_test, targets_test = [], []
        for image, target in test_images_and_targets:
            samples_test.append(image)
            targets_test.append(idx_to_class[target])

        train_df = pd.DataFrame({"samples": samples_train, "targets": targets_train})
        train_df["split"] = "train"
        val_df = pd.DataFrame({"samples": samples_val, "targets": targets_val})
        val_df["split"] = "val"
        test_df = pd.DataFrame({"samples": samples_test, "targets": targets_test})
        test_df["split"] = "test"
        self.data = pd.concat([train_df, val_df, test_df], axis=0)


class ImagenetteSSLDataModule(ImagenetteClassificationDataModule, SSLDataModule):
    """Initializes the SSL data module for Imagenette dataset."""

    def __init__(
        self,
        *args: Any,
        name="imagenette_ssl",
        **kwargs: Any,
    ):
        super().__init__(*args, name=name, **kwargs)  # type: ignore[misc]



# pylint: disable=unsubscriptable-object
from __future__ import annotations

from typing import Any

import numpy as np
import torch
from sklearn.model_selection import train_test_split
from torch.utils.data import DataLoader

from quadra.datamodules.classification import ClassificationDataModule
from quadra.datasets import TwoAugmentationDataset, TwoSetAugmentationDataset
from quadra.utils import utils

log = utils.get_logger(__name__)


class SSLDataModule(ClassificationDataModule):
    """Base class for all data modules for self supervised learning data modules.

    Args:
        data_path: Path to the data main folder.
        augmentation_dataset: Augmentation dataset
            for training dataset.
        name: The name for the data module. Defaults to  "ssl_datamodule".
        split_validation: Whether to split the validation set if . Defaults to True.
        **kwargs: The keyword arguments for the classification data module. Defaults to None.
    """

    def __init__(
        self,
        data_path: str,
        augmentation_dataset: TwoAugmentationDataset | TwoSetAugmentationDataset,
        name: str = "ssl_datamodule",
        split_validation: bool = True,
        **kwargs: Any,
    ):
        super().__init__(
            data_path=data_path,
            name=name,
            **kwargs,
        )
        self.augmentation_dataset = augmentation_dataset
        self.classifier_train_dataset: torch.utils.data.Dataset | None = None
        self.split_validation = split_validation

    def setup(self, stage: str | None = None) -> None:
        """Setup data module based on stages of training."""
        if stage == "fit":
            self.train_dataset = self.dataset(
                samples=self.train_data["samples"].tolist(),
                targets=self.train_data["targets"].tolist(),
                transform=self.train_transform,
            )

            if np.unique(self.train_data["targets"]).shape[0] > 1 and not self.split_validation:
                self.classifier_train_dataset = self.dataset(
                    samples=self.train_data["samples"].tolist(),
                    targets=self.train_data["targets"].tolist(),
                    transform=self.val_transform,
                )
                self.val_dataset = self.dataset(
                    samples=self.val_data["samples"].tolist(),
                    targets=self.val_data["targets"].tolist(),
                    transform=self.val_transform,
                )
            else:
                train_classifier_samples, val_samples, train_classifier_targets, val_targets = train_test_split(
                    self.val_data["samples"],
                    self.val_data["targets"],
                    test_size=0.3,
                    random_state=self.seed,
                    stratify=self.val_data["targets"],
                )

                self.classifier_train_dataset = self.dataset(
                    samples=train_classifier_samples,
                    targets=train_classifier_targets,
                    transform=self.test_transform,
                )

                self.val_dataset = self.dataset(
                    samples=val_samples,
                    targets=val_targets,
                    transform=self.val_transform,
                )

                log.warning(
                    "The training set contains only one class and cannot be used to train a classifier. To overcome "
                    "this issue 70% of the validation set is used to train the classifier. The remaining will be used "
                    "as standard validation. To disable this behaviour set the `split_validation` parameter to False."
                )
                self._check_train_dataset_config()
        if stage == "test":
            self.test_dataset = self.dataset(
                samples=self.test_data["samples"].tolist(),
                targets=self.test_data["targets"].tolist(),
                transform=self.test_transform,
            )

    def _check_train_dataset_config(self):
        """Check if train dataset is configured correctly."""
        if self.train_dataset is None:
            raise ValueError("Train dataset is not initialized")
        if self.augmentation_dataset is None:
            raise ValueError("Augmentation dataset is not initialized")
        if self.train_dataset.transform is not None:
            log.warning("Train dataset transform is not None. It will be applied before SSL augmentations")

    def train_dataloader(self) -> DataLoader:
        """Returns train dataloader."""
        if not isinstance(self.train_dataset, torch.utils.data.Dataset):
            raise ValueError("Train dataset is not a subclass of `torch.utils.data.Dataset`")
        self.augmentation_dataset.dataset = self.train_dataset
        loader = DataLoader(
            self.augmentation_dataset,
            batch_size=self.batch_size,
            shuffle=True,
            num_workers=self.num_workers,
            drop_last=False,
            pin_memory=True,
            persistent_workers=self.num_workers > 0,
        )
        return loader

    def classifier_train_dataloader(self) -> DataLoader:
        """Returns classifier train dataloader."""
        if self.classifier_train_dataset is None:
            raise ValueError("Classifier train dataset is not initialized")

        loader = DataLoader(
            self.classifier_train_dataset,
            batch_size=self.batch_size,
            shuffle=True,
            num_workers=self.num_workers,
            drop_last=False,
            pin_memory=True,
            persistent_workers=self.num_workers > 0,
        )
        return loader



from __future__ import annotations

import os
import random
from collections.abc import Callable

import cv2
import h5py
import numpy as np
from torch.utils.data import Dataset

from quadra.utils import utils
from quadra.utils.imaging import keep_aspect_ratio_resize
from quadra.utils.patch.dataset import compute_safe_patch_range, trisample

log = utils.get_logger(__name__)


class PatchSklearnClassificationTrainDataset(Dataset):
    """Dataset used for patch sampling, it expects samples to be paths to h5 files containing all the required
    information for patch sampling from images.

    Args:
        data_path: base path to the dataset
        samples: Paths to h5 files
        targets: Labels associated with each sample
        class_to_idx: Mapping between class and corresponding index
        resize: Whether to perform an aspect ratio resize of the patch before the transformations
        transform: Optional function applied to the image
        rgb: if False, image will be converted in grayscale
        channel: 1 or 3. If rgb is True, then channel will be set at 3.
        balance_classes: if True, the dataset will be balanced by duplicating samples of the minority class
    """

    def __init__(
        self,
        data_path: str,
        samples: list[str],
        targets: list[str | int],
        class_to_idx: dict | None = None,
        resize: int | None = None,
        transform: Callable | None = None,
        rgb: bool = True,
        channel: int = 3,
        balance_classes: bool = False,
    ):
        super().__init__()

        # Keep-Aspect-Ratio resize
        self.resize = resize
        self.data_path = data_path

        if balance_classes:
            samples_array = np.array(samples)
            targets_array = np.array(targets)
            samples_to_use: list[str] = []
            targets_to_use: list[str | int] = []

            cls, counts = np.unique(targets_array, return_counts=True)
            max_count = np.max(counts)
            for cl, count in zip(cls, counts, strict=False):
                idx_to_pick = list(np.where(targets_array == cl)[0])

                if count < max_count:
                    idx_to_pick += random.choices(idx_to_pick, k=max_count - count)

                samples_to_use.extend(samples_array[idx_to_pick])
                targets_to_use.extend(targets_array[idx_to_pick])
        else:
            samples_to_use = samples
            targets_to_use = targets

        # Data
        self.x = np.array(samples_to_use)
        self.y = np.array(targets_to_use)

        if class_to_idx is None:
            unique_targets = np.unique(targets_to_use)
            class_to_idx = {c: i for i, c in enumerate(unique_targets)}

        self.class_to_idx = class_to_idx
        self.idx_to_class = {v: k for k, v in class_to_idx.items()}

        self.samples = [
            (path, self.class_to_idx[self.y[i]] if self.y[i] is not None else None) for i, path in enumerate(self.x)
        ]

        self.rgb = rgb
        self.channel = 3 if rgb else channel

        self.transform = transform

    def __getitem__(self, idx) -> tuple[np.ndarray, np.ndarray]:
        path, y = self.samples[idx]

        h5_file = h5py.File(path)
        x = cv2.imread(os.path.join(self.data_path, h5_file["img_path"][()].decode("utf-8")))

        weights = h5_file["triangles_weights"][()]
        patch_size = h5_file["patch_size"][()]

        if weights.shape[0] == 0:  # pylint: disable=no-member
            # If the image is completely good sample a point anywhere
            patch_y = np.random.randint(0, x.shape[0] + 1)
            patch_x = np.random.randint(0, x.shape[1] + 1)
        else:
            random_triangle = np.random.choice(weights.shape[0], p=weights)
            [patch_y, patch_x] = trisample(h5_file["triangles"][random_triangle])

        h5_file.close()

        # If the patch is outside the image reduce the exceeding area by taking more patch from the inner area
        [y_left, y_right] = compute_safe_patch_range(patch_y, patch_size[0], x.shape[0])
        [x_left, x_right] = compute_safe_patch_range(patch_x, patch_size[1], x.shape[1])

        x = x[patch_y - y_left : patch_y + y_right, patch_x - x_left : patch_x + x_right]

        if self.rgb:
            x = cv2.cvtColor(x, cv2.COLOR_BGR2RGB)
        else:
            x = cv2.cvtColor(x, cv2.COLOR_BGR2GRAY)
            x = cv2.cvtColor(x, cv2.COLOR_GRAY2RGB)

        if self.channel == 1:
            x = x[:, :, 0]

        # Resize keeping aspect ratio
        if self.resize:
            x = keep_aspect_ratio_resize(x, self.resize)

        if self.transform:
            aug = self.transform(image=x)
            x = aug["image"]

        return x, y

    def __len__(self):
        return len(self.samples)



from __future__ import annotations

import inspect
from collections.abc import Sequence
from typing import Any

import torch
from torch import nn

from quadra.utils.logger import get_logger

log = get_logger(__name__)


class ModelSignatureWrapper(nn.Module):
    """Model wrapper used to retrieve input shape. It can be used as a decorator of nn.Module, the first call to the
    forward method will retrieve the input shape and store it in the input_shapes attribute.
    It will also save the model summary in a file called model_summary.txt in the current working directory.
    """

    def __init__(self, model: nn.Module):
        super().__init__()
        self.instance = model
        self.input_shapes: Any = None
        self.disable = False

        if isinstance(self.instance, ModelSignatureWrapper):
            # Handle nested ModelSignatureWrapper
            self.input_shapes = self.instance.input_shapes
            self.instance = self.instance.instance

    def forward(self, *args: Any, **kwargs: Any) -> torch.Tensor:
        """Retrieve the input shape and forward the model, if the input shape is already retrieved it will just forward
        the model.
        """
        if self.input_shapes is None and not self.disable:
            try:
                self.input_shapes = self._get_input_shapes(*args, **kwargs)
            except Exception:
                log.warning(
                    "Failed to retrieve input shapes after forward! To export the model you'll need to "
                    "provide the input shapes manually setting the config.export.input_shapes parameter! "
                    "Alternatively you could try to use a forward with supported input types (and their compositions) "
                    "(list, tuple, dict, tensors)."
                )
                self.disable = True

        return self.instance.forward(*args, **kwargs)

    def to(self, *args, **kwargs):
        """Handle calls to to method returning the underlying model."""
        self.instance = self.instance.to(*args, **kwargs)

        return self

    def half(self, *args, **kwargs):
        """Handle calls to to method returning the underlying model."""
        self.instance = self.instance.half(*args, **kwargs)

        return self

    def cpu(self, *args, **kwargs):
        """Handle calls to to method returning the underlying model."""
        self.instance = self.instance.cpu(*args, **kwargs)

        return self

    def _get_input_shapes(self, *args: Any, **kwargs: Any) -> list[Any]:
        """Retrieve the input shapes from the input. Inputs will be in the same order as the forward method
        signature.
        """
        input_shapes = []

        for arg in args:
            input_shapes.append(self._get_input_shape(arg))

        if isinstance(self.instance.forward, torch.ScriptMethod):
            # Handle torchscript backbones
            for i, argument in enumerate(self.instance.forward.schema.arguments):  # type: ignore[attr-defined]
                if i < (len(args) + 1):  # +1 for self
                    continue

                if argument.name == "self":
                    continue

                if argument.name in kwargs:
                    input_shapes.append(self._get_input_shape(kwargs[argument.name]))
                else:
                    # Retrieve the default value
                    input_shapes.append(self._get_input_shape(argument.default_value))
        else:
            signature = inspect.signature(self.instance.forward)

            for i, key in enumerate(signature.parameters.keys()):
                if i < len(args):
                    continue

                if key in kwargs:
                    input_shapes.append(self._get_input_shape(kwargs[key]))
                else:
                    # Retrieve the default value
                    input_shapes.append(self._get_input_shape(signature.parameters[key].default))

        return input_shapes

    def _get_input_shape(self, inp: Sequence | torch.Tensor) -> list[Any] | tuple[Any, ...] | dict[str, Any]:
        """Recursive function to retrieve the input shapes."""
        if isinstance(inp, list):
            return [self._get_input_shape(i) for i in inp]

        if isinstance(inp, tuple):
            return tuple(self._get_input_shape(i) for i in inp)

        if isinstance(inp, torch.Tensor):
            return tuple(inp.shape[1:])

        if isinstance(inp, dict):
            return {k: self._get_input_shape(v) for k, v in inp.items()}

        raise ValueError(f"Input type {type(inp)} not supported")

    def __getattr__(self, name: str) -> torch.Tensor | nn.Module:
        if name in ["instance", "input_shapes"]:
            return self.__dict__[name]

        return getattr(self.__dict__["instance"], name)

    def __setattr__(self, name: str, value: torch.Tensor | nn.Module) -> None:
        if name in ["instance", "input_shapes"]:
            self.__dict__[name] = value
        else:
            setattr(self.instance, name, value)

    def __getattribute__(self, __name: str) -> Any:
        if __name in [
            "instance",
            "input_shapes",
            "__dict__",
            "forward",
            "_get_input_shapes",
            "_get_input_shape",
            "to",
            "half",
            "cpu",
            "call_super_init",
            "_call_impl",
            "_compiled_call_impl",
        ]:
            return super().__getattribute__(__name)

        return getattr(self.instance, __name)



from .dataset import generate_patch_dataset, get_image_mask_association
from .metrics import compute_patch_metrics, reconstruct_patch
from .model import RleEncoder, save_classification_result
from .visualization import plot_patch_reconstruction, plot_patch_results

__all__ = [
    "generate_patch_dataset",
    "reconstruct_patch",
    "save_classification_result",
    "plot_patch_reconstruction",
    "plot_patch_results",
    "get_image_mask_association",
    "compute_patch_metrics",
    "RleEncoder",
]



from __future__ import annotations

import random
from collections.abc import Iterable
from enum import Enum

import albumentations as A
import numpy as np
from torch.utils.data import Dataset


class AugmentationStrategy(Enum):
    """Augmentation Strategy for TwoAugmentationDataset."""

    SAME_IMAGE = 1
    SAME_CLASS = 2


class TwoAugmentationDataset(Dataset):
    """Two Image Augmentation Dataset for using in self-supervised learning.

    Args:
        dataset: A torch Dataset object
        transform: albumentation transformations for each image.
            If you use single transformation, it will be applied to both images.
            If you use tuple, it will be applied to first image and second image separately.
        strategy: Defaults to AugmentationStrategy.SAME_IMAGE.
    """

    def __init__(
        self,
        dataset: Dataset,
        transform: A.Compose | tuple[A.Compose, A.Compose],
        strategy: AugmentationStrategy = AugmentationStrategy.SAME_IMAGE,
    ):
        self.dataset = dataset
        self.transform = transform
        self.stategy = strategy
        if isinstance(transform, Iterable) and not isinstance(transform, str) and len(set(transform)) != 2:
            raise ValueError("transform must be an Iterable of length 2")

    def __getitem__(self, index):
        image1, target = self.dataset[index]

        if self.stategy == AugmentationStrategy.SAME_IMAGE:
            image2 = image1
        elif self.stategy == AugmentationStrategy.SAME_CLASS:
            positive_pair_idx = random.choice(np.where(self.dataset.y == target)[0])
            image2, _ = self.dataset[positive_pair_idx]
        else:
            raise ValueError("Unknown strategy")

        if isinstance(self.transform, Iterable):
            image1 = self.transform[0](image=image1)["image"]
            image2 = self.transform[1](image=image2)["image"]
        else:
            image1 = self.transform(image=image1)["image"]
            image2 = self.transform(image=image2)["image"]

        return [image1, image2], target

    def __len__(self):
        return len(self.dataset)


class TwoSetAugmentationDataset(Dataset):
    """Two Set Augmentation Dataset for using in self-supervised learning (DINO).

    Args:
        dataset: Base dataset
        global_transforms: Global transformations for each image.
        local_transform: Local transformations for each image.
        num_local_transforms: Number of local transformations to apply. In total you will have
            two + num_local_transforms transformations for each image. First element of the array will always
            return the original image.

    Example:
        >>> `images[0] = global_transform[0](original_image)`
        >>> `images[1] = global_transform[1](original_image)`
        >>> `images[2:] = local_transform(s)(original_image)`
    """

    def __init__(
        self,
        dataset: Dataset,
        global_transforms: tuple[A.Compose, A.Compose],
        local_transform: A.Compose,
        num_local_transforms: int,
    ):
        self.dataset = dataset
        self.global_transforms = global_transforms
        self.local_transform = local_transform
        self.num_local_transforms = num_local_transforms

        if num_local_transforms < 1:
            raise ValueError("num_local_transforms must be greater than 0")

    def __getitem__(self, index):
        original_image, target = self.dataset[index]
        global_outputs = []
        local_outputs = []
        for global_transform in self.global_transforms:
            global_outputs.append(global_transform(image=original_image)["image"])
        for _ in range(self.num_local_transforms):
            local_outputs.append(self.local_transform(image=original_image)["image"])
        all_outputs = global_outputs + local_outputs
        return all_outputs, target

    def __len__(self):
        return len(self.dataset)



from __future__ import annotations

try:
    from mlflow.models import infer_signature  # noqa
    from mlflow.models.signature import ModelSignature  # noqa

    MLFLOW_AVAILABLE = True
except ImportError:
    MLFLOW_AVAILABLE = False

from collections.abc import Sequence
from typing import Any

import numpy as np
import torch
from pytorch_lightning import Trainer
from pytorch_lightning.loggers import MLFlowLogger

from quadra.models.evaluation import BaseEvaluationModel


@torch.inference_mode()
def infer_signature_model(model: BaseEvaluationModel, data: list[Any]) -> ModelSignature | None:
    """Infer input and output signature for a PyTorch/Torchscript model."""
    model = model.eval()
    model_output = model(*data)

    try:
        output_signature = infer_signature_input(model_output)

        if len(data) == 1:
            signature_input = infer_signature_input(data[0])
        else:
            signature_input = infer_signature_input(data)
    except ValueError:
        # TODO: Solve circular import as it is not possible to import get_logger right now
        # log.warning("Unable to infer signature for model output type %s", type(model_output))
        return None

    return infer_signature(signature_input, output_signature)


def infer_signature_input(input_tensor: Any) -> Any:
    """Recursively infer the signature input format to pass to mlflow.models.infer_signature.

    Raises:
        ValueError: If the input type is not supported or when nested dicts or sequences are encountered.
    """
    signature: dict[str, Any] | np.ndarray
    if isinstance(input_tensor, Sequence):
        # Mlflow currently does not support sequence outputs, so we use a dict instead
        signature = {}
        for i, x in enumerate(input_tensor):
            if isinstance(x, Sequence):
                # Nested signature is currently not supported by mlflow
                raise ValueError("Nested sequences are not supported")
                # TODO: Enable this once mlflow supports nested signatures
                # signature[f"output_{i}"] = {f"output_{j}": infer_signature_torch(y) for j, y in enumerate(x)}
            if isinstance(x, dict):
                # Nested dicts are not supported
                raise ValueError("Nested dicts are not supported")

            signature[f"output_{i}"] = infer_signature_input(x)
    elif isinstance(input_tensor, torch.Tensor):
        signature = input_tensor.cpu().numpy()
    elif isinstance(input_tensor, dict):
        signature = {}
        for k, v in input_tensor.items():
            if isinstance(v, dict):
                # Nested dicts are not supported
                raise ValueError("Nested dicts are not supported")
            if isinstance(v, Sequence):
                # Nested signature is currently not supported by mlflow
                raise ValueError("Nested sequences are not supported")

            signature[k] = infer_signature_input(v)
    else:
        raise ValueError(f"Unable to infer signature for model output type {type(input_tensor)}")

    return signature


def get_mlflow_logger(trainer: Trainer) -> MLFlowLogger | None:
    """Safely get Mlflow logger from Trainer loggers.

    Args:
        trainer: Pytorch Lightning trainer.

    Returns:
        An mlflow logger if available, else None.
    """
    if isinstance(trainer.logger, MLFlowLogger):
        return trainer.logger

    if isinstance(trainer.logger, list):
        for logger in trainer.logger:
            if isinstance(logger, MLFlowLogger):
                return logger

    return None



from __future__ import annotations

import os
import uuid
from copy import deepcopy
from typing import Any

import pytorch_lightning as pl
from pytorch_lightning.callbacks import Callback
from pytorch_lightning.callbacks.batch_size_finder import BatchSizeFinder as LightningBatchSizeFinder
from pytorch_lightning.utilities import rank_zero_only
from pytorch_lightning.utilities.exceptions import _TunerExitException
from pytorch_lightning.utilities.memory import garbage_collection_cuda, is_oom_error
from pytorch_lightning.utilities.parsing import lightning_getattr, lightning_setattr
from pytorch_lightning.utilities.rank_zero import rank_zero_info, rank_zero_warn
from torch import nn

from quadra.utils.utils import get_logger

log = get_logger(__name__)

# pylint: disable=protected-access


def _scale_batch_size(
    trainer: pl.Trainer,
    mode: str = "power",
    steps_per_trial: int = 3,
    init_val: int = 2,
    max_trials: int = 25,
    batch_arg_name: str = "batch_size",
) -> int | None:
    """Iteratively try to find the largest batch size for a given model that does not give an out of memory (OOM)
    error.

    Args:
        trainer: A Trainer instance.
        mode: Search strategy to update the batch size:

            - ``'power'``: Keep multiplying the batch size by 2, until we get an OOM error.
            - ``'binsearch'``: Initially keep multiplying by 2 and after encountering an OOM error
                do a binary search between the last successful batch size and the batch size that failed.

        steps_per_trial: number of steps to run with a given batch size.
            Ideally 1 should be enough to test if an OOM error occurs,
            however in practise a few are needed
        init_val: initial batch size to start the search with
        max_trials: max number of increases in batch size done before
            algorithm is terminated
        batch_arg_name: name of the attribute that stores the batch size.
            It is expected that the user has provided a model or datamodule that has a hyperparameter
            with that name. We will look for this attribute name in the following places

            - ``model``
            - ``model.hparams``
            - ``trainer.datamodule`` (the datamodule passed to the tune method)

    """
    if trainer.fast_dev_run:  # type: ignore[attr-defined]
        rank_zero_warn("Skipping batch size scaler since `fast_dev_run` is enabled.")
        return None

    # Save initial model, that is loaded after batch size is found
    ckpt_path = os.path.join(trainer.default_root_dir, f".scale_batch_size_{uuid.uuid4()}.ckpt")
    trainer.save_checkpoint(ckpt_path)

    # Arguments we adjust during the batch size finder, save for restoring
    params = __scale_batch_dump_params(trainer)

    # Set to values that are required by the algorithm
    __scale_batch_reset_params(trainer, steps_per_trial)

    if trainer.progress_bar_callback:
        trainer.progress_bar_callback.disable()

    lightning_setattr(trainer.lightning_module, batch_arg_name, init_val)

    if mode == "power":
        new_size = _run_power_scaling(trainer, init_val, batch_arg_name, max_trials, params)
    elif mode == "binsearch":
        new_size = _run_binary_scaling(trainer, init_val, batch_arg_name, max_trials, params)
    else:
        raise ValueError(f"Unknown mode {mode}")

    garbage_collection_cuda()

    log.info("Finished batch size finder, will continue with full run using batch size %d", new_size)

    __scale_batch_restore_params(trainer, params)

    if trainer.progress_bar_callback:
        trainer.progress_bar_callback.enable()

    trainer._checkpoint_connector.restore(ckpt_path)
    trainer.strategy.remove_checkpoint(ckpt_path)

    return new_size


def __scale_batch_dump_params(trainer: pl.Trainer) -> dict[str, Any]:
    """Dump the parameters that need to be reset after the batch size finder.."""
    dumped_params = {
        "loggers": trainer.loggers,
        "callbacks": trainer.callbacks,  # type: ignore[attr-defined]
    }
    loop = trainer._active_loop
    assert loop is not None
    if isinstance(loop, pl.loops._FitLoop):
        dumped_params["max_steps"] = trainer.max_steps
        dumped_params["limit_train_batches"] = trainer.limit_train_batches
        dumped_params["limit_val_batches"] = trainer.limit_val_batches
    elif isinstance(loop, pl.loops._EvaluationLoop):
        stage = trainer.state.stage
        assert stage is not None
        dumped_params["limit_eval_batches"] = getattr(trainer, f"limit_{stage.dataloader_prefix}_batches")
        dumped_params["loop_verbose"] = loop.verbose

    dumped_params["loop_state_dict"] = deepcopy(loop.state_dict())
    return dumped_params


def __scale_batch_reset_params(trainer: pl.Trainer, steps_per_trial: int) -> None:
    """Reset the parameters that need to be reset after the batch size finder."""
    from pytorch_lightning.loggers.logger import DummyLogger  # pylint: disable=import-outside-toplevel

    trainer.logger = DummyLogger() if trainer.logger is not None else None
    trainer.callbacks = []  # type: ignore[attr-defined]

    loop = trainer._active_loop
    assert loop is not None
    if isinstance(loop, pl.loops._FitLoop):
        trainer.limit_train_batches = 1.0
        trainer.limit_val_batches = steps_per_trial
        trainer.fit_loop.epoch_loop.max_steps = steps_per_trial
    elif isinstance(loop, pl.loops._EvaluationLoop):
        stage = trainer.state.stage
        assert stage is not None
        setattr(trainer, f"limit_{stage.dataloader_prefix}_batches", steps_per_trial)
        loop.verbose = False


def __scale_batch_restore_params(trainer: pl.Trainer, params: dict[str, Any]) -> None:
    """Restore the parameters that need to be reset after the batch size finder."""
    # TODO: There are more states that needs to be reset (#4512 and #4870)
    trainer.loggers = params["loggers"]
    trainer.callbacks = params["callbacks"]  # type: ignore[attr-defined]

    loop = trainer._active_loop
    assert loop is not None
    if isinstance(loop, pl.loops._FitLoop):
        loop.epoch_loop.max_steps = params["max_steps"]
        trainer.limit_train_batches = params["limit_train_batches"]
        trainer.limit_val_batches = params["limit_val_batches"]
    elif isinstance(loop, pl.loops._EvaluationLoop):
        stage = trainer.state.stage
        assert stage is not None
        setattr(trainer, f"limit_{stage.dataloader_prefix}_batches", params["limit_eval_batches"])

    loop.load_state_dict(deepcopy(params["loop_state_dict"]))
    loop.restarting = False
    if isinstance(loop, pl.loops._EvaluationLoop) and "loop_verbose" in params:
        loop.verbose = params["loop_verbose"]

    # make sure the loop's state is reset
    _reset_dataloaders(trainer)
    loop.reset()


def _run_power_scaling(
    trainer: pl.Trainer,
    new_size: int,
    batch_arg_name: str,
    max_trials: int,
    params: dict[str, Any],
) -> int:
    """Batch scaling mode where the size is doubled at each iteration until an OOM error is encountered."""
    # this flag is used to determine whether the previously scaled batch size, right before OOM, was a success or not
    # if it was we exit, else we continue downscaling in case we haven't encountered a single optimal batch size
    any_success = False
    # In the original
    for i in range(max_trials):
        garbage_collection_cuda()

        # reset after each try
        _reset_progress(trainer)

        try:
            if i == 0:
                rank_zero_info(f"Starting batch size finder with batch size {new_size}")
                new_size, _ = _adjust_batch_size(trainer, batch_arg_name, factor=1.0, desc=None)
                changed = True
            else:
                new_size, changed = _adjust_batch_size(trainer, batch_arg_name, factor=2.0, desc="succeeded")
            # Force the train dataloader to reset as the batch size has changed
            _reset_dataloaders(trainer)
            _try_loop_run(trainer, params)

            any_success = True

            # In the original lightning implementation this is done before _reset_dataloaders
            # As such the batch size is not checked for the last iteration!!!
            if not changed:
                break
        except RuntimeError as exception:
            if is_oom_error(exception):
                # If we fail in power mode, half the size and return
                garbage_collection_cuda()
                if any_success:
                    # In the original lightning code there's a line that doesn't halve the size properly if batch_size
                    # is bigger than the dataset length
                    rank_zero_info(f"Batch size {new_size} failed, using batch size {new_size // 2}")
                    new_size = new_size // 2
                    lightning_setattr(trainer.lightning_module, batch_arg_name, new_size)
                else:
                    # In this case it means the first iteration will fail already, probably due to a way to big
                    # initial batch size, since the next iteration will start from (new_size // 2) * 2, which is the
                    # same divide by 4 instead and retry
                    rank_zero_info(f"Batch size {new_size} failed at first iteration, using batch size {new_size // 4}")
                    new_size = new_size // 4
                    lightning_setattr(trainer.lightning_module, batch_arg_name, new_size)

                # Force the train dataloader to reset as the batch size has changed
                _reset_dataloaders(trainer)
                if any_success:
                    break
            else:
                raise  # some other error not memory related

    return new_size


def _run_binary_scaling(
    trainer: pl.Trainer,
    new_size: int,
    batch_arg_name: str,
    max_trials: int,
    params: dict[str, Any],
) -> int:
    """Batch scaling mode where the size is initially is doubled at each iteration until an OOM error is encountered.

    Hereafter, the batch size is further refined using a binary search

    """
    low = 1
    high = None
    count = 0
    while True:
        garbage_collection_cuda()

        # reset after each try
        _reset_progress(trainer)

        try:
            # run loop
            _try_loop_run(trainer, params)
            count += 1
            if count > max_trials:
                break
            # Double in size
            low = new_size
            if high:
                if high - low <= 1:
                    break
                midval = (high + low) // 2
                new_size, changed = _adjust_batch_size(trainer, batch_arg_name, value=midval, desc="succeeded")
            else:
                new_size, changed = _adjust_batch_size(trainer, batch_arg_name, factor=2.0, desc="succeeded")

            if not changed:
                break

            # Force the train dataloader to reset as the batch size has changed
            _reset_dataloaders(trainer)

        except RuntimeError as exception:
            # Only these errors should trigger an adjustment
            if is_oom_error(exception):
                # If we fail in power mode, half the size and return
                garbage_collection_cuda()

                high = new_size
                midval = (high + low) // 2
                new_size, _ = _adjust_batch_size(trainer, batch_arg_name, value=midval, desc="failed")

                # Force the train dataloader to reset as the batch size has changed
                _reset_dataloaders(trainer)

                if high - low <= 1:
                    break
            else:
                raise  # some other error not memory related

    return new_size


def _adjust_batch_size(
    trainer: pl.Trainer,
    batch_arg_name: str = "batch_size",
    factor: float = 1.0,
    value: int | None = None,
    desc: str | None = None,
) -> tuple[int, bool]:
    """Helper function for adjusting the batch size.

    Args:
        trainer: instance of pytorch_lightning.Trainer
        batch_arg_name: name of the attribute that stores the batch size
        factor: value which the old batch size is multiplied by to get the
            new batch size
        value: if a value is given, will override the batch size with this value.
            Note that the value of `factor` will not have an effect in this case
        desc: either ``"succeeded"`` or ``"failed"``. Used purely for logging

    Returns:
        The new batch size for the next trial and a bool that signals whether the
        new value is different than the previous batch size.

    """
    model = trainer.lightning_module
    batch_size = lightning_getattr(model, batch_arg_name)
    assert batch_size is not None

    loop = trainer._active_loop
    assert loop is not None
    loop.setup_data()
    combined_loader = loop._combined_loader
    assert combined_loader is not None
    try:
        combined_dataset_length = combined_loader._dataset_length()
        if batch_size >= combined_dataset_length:
            rank_zero_info(f"The batch size {batch_size} is greater or equal than the length of your dataset.")
            return batch_size, False
    except NotImplementedError:
        # all datasets are iterable style
        pass

    new_size = value if value is not None else int(batch_size * factor)
    if desc:
        rank_zero_info(f"Batch size {batch_size} {desc}, trying batch size {new_size}")
    changed = new_size != batch_size
    lightning_setattr(model, batch_arg_name, new_size)

    return new_size, changed


def _reset_dataloaders(trainer: pl.Trainer) -> None:
    """Reset the dataloaders to force a reload."""
    loop = trainer._active_loop
    assert loop is not None
    loop._combined_loader = None  # force a reload
    loop.setup_data()
    if isinstance(loop, pl.loops._FitLoop):
        loop.epoch_loop.val_loop._combined_loader = None
        loop.epoch_loop.val_loop.setup_data()


def _try_loop_run(trainer: pl.Trainer, params: dict[str, Any]) -> None:
    """Try to run the loop with the current batch size."""
    loop = trainer._active_loop
    assert loop is not None
    loop.load_state_dict(deepcopy(params["loop_state_dict"]))
    loop.restarting = False
    loop.run()


def _reset_progress(trainer: pl.Trainer) -> None:
    """Reset the progress of the trainer."""
    if trainer.lightning_module.automatic_optimization:
        trainer.fit_loop.epoch_loop.automatic_optimization.optim_progress.reset()
    else:
        trainer.fit_loop.epoch_loop.manual_optimization.optim_step_progress.reset()

    trainer.fit_loop.epoch_progress.reset()


# Most of the code above is copied from the original lightning implementation since almost everything is private


class LightningTrainerBaseSetup(Callback):
    """Custom callback used to setup a lightning trainer with default options.

    Args:
        log_every_n_steps: Default value for trainer.log_every_n_steps if the dataloader is too small.
    """

    def __init__(self, log_every_n_steps: int = 1) -> None:
        self.log_every_n_steps = log_every_n_steps

    @rank_zero_only
    def on_fit_start(self, trainer: pl.Trainer, pl_module: pl.LightningModule) -> None:
        """Called on every stage."""
        if not hasattr(trainer, "datamodule") or not hasattr(trainer, "log_every_n_steps"):
            raise ValueError("Trainer must have a datamodule and log_every_n_steps attribute.")

        len_train_dataloader = len(trainer.datamodule.train_dataloader())
        if len_train_dataloader <= trainer.log_every_n_steps:
            if len_train_dataloader > self.log_every_n_steps:
                trainer.log_every_n_steps = self.log_every_n_steps
                log.info("`trainer.log_every_n_steps` is too high, setting it to %d", self.log_every_n_steps)
            else:
                trainer.log_every_n_steps = 1
                log.warning(
                    "The default log_every_n_steps %d is too high given the datamodule lenght %d, fallback to 1",
                    self.log_every_n_steps,
                    len_train_dataloader,
                )


class BatchSizeFinder(LightningBatchSizeFinder):
    """Batch size finder setting the proper model training status as the current one from lightning seems bugged.
    It also allows to skip some batch size finding steps.

    Args:
        find_train_batch_size: Whether to find the training batch size.
        find_validation_batch_size: Whether to find the validation batch size.
        find_test_batch_size: Whether to find the test batch size.
        find_predict_batch_size: Whether to find the predict batch size.
        mode: The mode to use for batch size finding. See `pytorch_lightning.callbacks.BatchSizeFinder` for more
            details.
        steps_per_trial: The number of steps per trial. See `pytorch_lightning.callbacks.BatchSizeFinder` for more
            details.
        init_val: The initial value for batch size. See `pytorch_lightning.callbacks.BatchSizeFinder` for more details.
        max_trials: The maximum number of trials. See `pytorch_lightning.callbacks.BatchSizeFinder` for more details.
        batch_arg_name: The name of the batch size argument. See `pytorch_lightning.callbacks.BatchSizeFinder` for more
            details.
    """

    def __init__(
        self,
        find_train_batch_size: bool = True,
        find_validation_batch_size: bool = False,
        find_test_batch_size: bool = False,
        find_predict_batch_size: bool = False,
        mode: str = "power",
        steps_per_trial: int = 3,
        init_val: int = 2,
        max_trials: int = 25,
        batch_arg_name: str = "batch_size",
    ) -> None:
        super().__init__(mode, steps_per_trial, init_val, max_trials, batch_arg_name)
        self.find_train_batch_size = find_train_batch_size
        self.find_validation_batch_size = find_validation_batch_size
        self.find_test_batch_size = find_test_batch_size
        self.find_predict_batch_size = find_predict_batch_size

    def on_fit_start(self, trainer: pl.Trainer, pl_module: pl.LightningModule) -> None:
        if not self.find_train_batch_size or trainer.state.stage is None:
            # If called during validation skip it as it will be triggered during on_validation_start
            return None

        if trainer.state.stage.value != "train":
            return None

        if not isinstance(pl_module.model, nn.Module):
            raise ValueError("The model must be a nn.Module")
        pl_module.model.train()

        return super().on_fit_start(trainer, pl_module)

    def on_validation_start(self, trainer: pl.Trainer, pl_module: pl.LightningModule) -> None:
        if not self.find_validation_batch_size:
            return None

        if not isinstance(pl_module.model, nn.Module):
            raise ValueError("The model must be a nn.Module")
        pl_module.model.eval()

        return super().on_validation_start(trainer, pl_module)

    def on_test_start(self, trainer: pl.Trainer, pl_module: pl.LightningModule) -> None:
        if not self.find_test_batch_size:
            return None

        if not isinstance(pl_module.model, nn.Module):
            raise ValueError("The model must be a nn.Module")
        pl_module.model.eval()

        return super().on_test_start(trainer, pl_module)

    def on_predict_start(self, trainer: pl.Trainer, pl_module: pl.LightningModule) -> None:
        if not self.find_predict_batch_size:
            return None

        if not isinstance(pl_module.model, nn.Module):
            raise ValueError("The model must be a nn.Module")
        pl_module.model.eval()

        return super().on_predict_start(trainer, pl_module)

    def scale_batch_size(self, trainer: pl.Trainer, pl_module: pl.LightningModule) -> None:
        """Scale the batch size."""
        new_size = _scale_batch_size(
            trainer,
            self._mode,
            self._steps_per_trial,
            self._init_val,
            self._max_trials,
            self._batch_arg_name,
        )

        self.optimal_batch_size = new_size
        if self._early_exit:
            raise _TunerExitException()



import math

import torch
import torch.nn.functional as F

from quadra.utils.utils import AllGatherSyncFunction


def simclr_loss(
    features1: torch.Tensor,
    features2: torch.Tensor,
    temperature: float = 1.0,
) -> torch.Tensor:
    """SimCLR loss described in https://arxiv.org/pdf/2002.05709.pdf.

    Args:
        temperature: optional temperature
        features1: First augmented features (i.e. T(features))
        features2: Second augmented features (i.e. T'(features))

    Returns:
        SimCLR loss
    """
    features1 = F.normalize(features1, dim=-1)
    features2 = F.normalize(features2, dim=-1)
    if torch.distributed.is_available() and torch.distributed.is_initialized():
        features1_dist = AllGatherSyncFunction.apply(features1)
        features2_dist = AllGatherSyncFunction.apply(features2)
    else:
        features1_dist = features1
        features2_dist = features2
    features = torch.cat([features1, features2], dim=0)  # [2B, d]
    features_dist = torch.cat([features1_dist, features2_dist], dim=0)  # [2B * DIST_SIZE, d]

    # Similarity matrix
    sim = torch.exp(torch.div(torch.mm(features, features_dist.t()), temperature))  # [2B, 2B * DIST_SIZE]

    # Negatives
    neg = sim.sum(dim=-1)

    # From each row, subtract e^(1/temp) to remove similarity measure for zi * zi, since
    # (zi^T * zi) / ||zi||^2 = 1
    row_sub = torch.full_like(neg, math.e ** (1 / temperature), device=neg.device)
    neg = torch.clamp(neg - row_sub, min=1e-6)  # clamp for numerical stability

    # Positive similarity, pos becomes [2 * batch_size]
    pos = torch.exp(torch.div(torch.sum(features1 * features2, dim=-1), temperature))
    pos = torch.cat([pos, pos], dim=0)

    loss = -torch.log(pos / (neg + 1e-6)).mean()
    return loss


class SimCLRLoss(torch.nn.Module):
    """SIMCLRloss module.

    Args:
        temperature: temperature of SIM loss.
    """

    def __init__(self, temperature: float = 1.0):
        super().__init__()
        self.temperature = temperature

    def forward(self, x1: torch.Tensor, x2: torch.Tensor) -> torch.Tensor:
        """Forward pass of the loss."""
        return simclr_loss(x1, x2, temperature=self.temperature)



from __future__ import annotations

import sklearn
import torch
from torch import nn

from quadra.modules.base import SSLModule


class SimCLR(SSLModule):
    """SIMCLR class.

    Args:
        model: Feature extractor as pytorch `torch.nn.Module`
        projection_mlp: projection head as
            pytorch `torch.nn.Module`
        criterion: SSL loss to be applied
        classifier: Standard sklearn classifier. Defaults to None.
        optimizer: optimizer of the training. If None a default Adam is used. Defaults to None.
        lr_scheduler: lr scheduler. If None a default ReduceLROnPlateau is used. Defaults to None.
        lr_scheduler_interval: interval at which the lr scheduler is updated. Defaults to "epoch".
    """

    def __init__(
        self,
        model: nn.Module,
        projection_mlp: nn.Module,
        criterion: torch.nn.Module,
        classifier: sklearn.base.ClassifierMixin | None = None,
        optimizer: torch.optim.Optimizer | None = None,
        lr_scheduler: object | None = None,
        lr_scheduler_interval: str | None = "epoch",
    ):
        super().__init__(
            model,
            criterion,
            classifier,
            optimizer,
            lr_scheduler,
            lr_scheduler_interval,
        )
        self.projection_mlp = projection_mlp

    def forward(self, x):
        x = self.model(x)
        x = self.projection_mlp(x)
        return x

    def training_step(
        self, batch: tuple[tuple[torch.Tensor, torch.Tensor], torch.Tensor], batch_idx: int
    ) -> torch.Tensor:
        """Args:
            batch: The batch of data
            batch_idx: The index of the batch.

        Returns:
            The computed loss
        """
        # pylint: disable=unused-argument
        (im_x, im_y), _ = batch
        emb_x = self(im_x)
        emb_y = self(im_y)
        loss = self.criterion(emb_x, emb_y)

        self.log(
            "loss",
            loss,
            on_epoch=True,
            on_step=True,
            logger=True,
            prog_bar=True,
        )
        return loss



from __future__ import annotations

import torch
from torch.nn import functional as F


def euclidean_dist(
    query: torch.Tensor,
    prototypes: torch.Tensor,
    sen: bool = True,
    eps_pos: float = 1.0,
    eps_neg: float = -1e-7,
    eps: float = 1e-7,
) -> torch.Tensor:
    """Compute euclidean distance between two tensors.
        SEN dissimilarity from https://www.ecva.net/papers/eccv_2020/papers_ECCV/papers/123680120.pdf
    Args:
        query: feature of the network
        prototypes: prototypes of the center
        sen: Sen dissimilarity flag
        eps_pos: similarity arg
        eps_neg: similarity arg
        eps: similarity arg.

    Returns:
        Euclidian loss

    """
    # query: (n_classes * n_query) x d
    # prototypes: n_classes x d
    n = query.size(0)
    m = prototypes.size(0)
    d = query.size(1)
    if d != prototypes.size(1):
        raise ValueError("query and prototypes size[1] should be equal")

    if sen:
        # SEN dissimilarity from https://www.ecva.net/papers/eccv_2020/papers_ECCV/papers/123680120.pdf
        norm_query = torch.linalg.norm(query, ord=2, dim=1)  # (n_classes * n_query) X 1
        norm_prototypes = torch.linalg.norm(prototypes, ord=2, dim=1)  # n_classes X 1

        # We have to compute (||z|| - ||c||)^2 between all query points w.r.t.
        # all support points

        # Replicate each single query norm value m times
        norm_query = norm_query.view(-1, 1).unsqueeze(1).expand(n, m, 1)
        # Replicate all prototypes norm values n times
        norm_prototypes = norm_prototypes.view(-1, 1).unsqueeze(0).expand(n, m, 1)
        norm_diff = torch.pow(norm_query - norm_prototypes, 2).squeeze(2)
        epsilon = torch.full((n, m), eps_neg).type_as(query)
        if eps_pos != eps_neg:
            # n_query = n // m
            # for i in range(m):
            #     epsilon[i * n_query : (i + 1) * n_query, i] = 1.0

            # Since query points with class i need to have a positive epsilon
            # whenever they refer to support point with class i and since
            # query and support points are ordered, we need to set:
            # the 1st column of the 1st n_query rows to eps_pos
            # the 2nd column of the 2nd n_query rows to eps_pos
            # and so on
            idxs = torch.eye(m, dtype=torch.bool).unsqueeze(1).expand(m, n // m, m).reshape(-1, m)
            epsilon[idxs] = eps_pos
        norm_diff = norm_diff * epsilon

    # Replicate each single query point value m times
    query = query.unsqueeze(1).expand(n, m, d)
    # Replicate all prototype points values n times
    prototypes = prototypes.unsqueeze(0).expand(n, m, d)

    norm = torch.pow(query - prototypes, 2).sum(2)
    if sen:
        return torch.sqrt(norm + norm_diff + eps)

    return norm


def prototypical_loss(
    coords: torch.Tensor,
    target: torch.Tensor,
    n_support: int,
    prototypes: torch.Tensor | None = None,
    sen: bool = True,
    eps_pos: float = 1.0,
    eps_neg: float = -1e-7,
):
    """Prototypical loss implementation.

    Inspired by https://github.com/jakesnell/prototypical-networks/blob/master/protonets/models/few_shot.py
        Compute the barycentres by averaging the features of n_support
        samples for each class in target, computes then the distances from each
        samples' features to each one of the barycentres, computes the
        log_probability for each n_query samples for each one of the current
        classes, of appartaining to a class c, loss and accuracy are then computed and returned.

    Args:
        coords: The model output for a batch of samples
        target: Ground truth for the above batch of samples
        n_support: Number of samples to keep in account when computing
            barycentres, for each one of the current classes
        prototypes: if not None, is used for classification
        sen: Sen dissimilarity flag
        eps_pos: Sen positive similarity arg
        eps_neg: Sen negative similarity arg
    """
    classes = torch.unique(target, sorted=True)
    n_classes = len(classes)
    n_query = len(torch.where(target == classes[0])[0]) - n_support

    # Check equality between classes and target with broadcasting:
    # class_idxs[i, j] = True iff classes[i] == target[j]
    class_idxs = classes.unsqueeze(1) == target
    if prototypes is None:
        # Get the prototypes as the mean of the support points,
        # ordered by class
        prototypes = torch.stack([coords[idx_list][:n_support] for idx_list in class_idxs]).mean(1)  # n_classes X d
    # Get query samples as the points NOT in the support set,
    # where, after .view(-1, d), one has that
    # the 1st n_query points refer to class 1
    # the 2nd n_query points refer to class 2
    # and so on
    query_samples = torch.stack([coords[idx_list][n_support:] for idx_list in class_idxs]).view(
        -1, prototypes.shape[-1]
    )  # (n_classes * n_query) X d
    # Get distances, where dists[i, j] is the distance between
    # query point i to support point j
    dists = euclidean_dist(
        query_samples, prototypes, sen=sen, eps_pos=eps_pos, eps_neg=eps_neg
    )  # (n_classes * n_query) X n_classes
    log_p_y = F.log_softmax(-dists, dim=1)
    log_p_y = log_p_y.view(n_classes, n_query, -1)  # n_classes X n_query X n_classes

    target_inds = torch.arange(0, n_classes).view(n_classes, 1, 1)
    # One solution is to use type_as(coords[0])
    target_inds = target_inds.type_as(coords)
    target_inds = target_inds.expand(n_classes, n_query, 1).long()

    # Since we need to backpropagate the log softmax of query points
    # of class i that refers to support of the same class for every i,
    # and since query and support are ordered we select:
    # from the 1st n_query X n_classes the 1st column
    # from the 2nd n_query X n_classes the 2st column
    # and so on
    loss_val = -log_p_y.gather(2, target_inds).squeeze().view(-1).mean()
    _, y_hat = log_p_y.max(2)
    acc_val = y_hat.eq(target_inds.squeeze()).float().mean()

    return loss_val, acc_val, prototypes



from __future__ import annotations

import json
import os
from typing import Any, cast

import hydra
import torch
from omegaconf import DictConfig, open_dict
from pytorch_lightning import LightningModule
from torch import nn
from torch.nn.functional import interpolate
from torch.utils.tensorboard import SummaryWriter
from tqdm import tqdm

from quadra.callbacks.scheduler import WarmupInit
from quadra.models.base import ModelSignatureWrapper
from quadra.models.evaluation import BaseEvaluationModel
from quadra.tasks.base import LightningTask, Task
from quadra.utils import utils
from quadra.utils.export import export_model, import_deployment_model

log = utils.get_logger(__name__)


class SSL(LightningTask):
    """SSL Task.

    Args:
        config: The experiment configuration
        checkpoint_path: The path to the checkpoint to load the model from Defaults to None
        report: Whether to create the report
        run_test: Whether to run final test
    """

    def __init__(
        self,
        config: DictConfig,
        run_test: bool = False,
        report: bool = False,
        checkpoint_path: str | None = None,
    ):
        super().__init__(
            config=config,
            checkpoint_path=checkpoint_path,
            run_test=run_test,
            report=report,
        )
        self._backbone: nn.Module
        self._optimizer: torch.optim.Optimizer
        self._lr_scheduler: torch.optim.lr_scheduler._LRScheduler
        self.export_folder = "deployment_model"

    def learnable_parameters(self) -> list[nn.Parameter]:
        """Get the learnable parameters."""
        raise NotImplementedError("This method must be implemented by the subclass")

    @property
    def optimizer(self) -> torch.optim.Optimizer:
        """Get the optimizer."""
        return self._optimizer

    @optimizer.setter
    def optimizer(self, optimizer_config: DictConfig) -> None:
        """Set the optimizer."""
        log.info("Instantiating optimizer <%s>", self.config.optimizer["_target_"])
        self._optimizer = hydra.utils.instantiate(optimizer_config, self.learnable_parameters())

    @property
    def scheduler(self) -> torch.optim.lr_scheduler._LRScheduler:
        """Get the scheduler."""
        return self._scheduler

    @scheduler.setter
    def scheduler(self, scheduler_config: DictConfig) -> None:
        log.info("Instantiating scheduler <%s>", scheduler_config["_target_"])
        if "CosineAnnealingWithLinearWarmUp" in self.config.scheduler["_target_"]:
            # This scheduler will be overwritten by the SSLCallback
            self._scheduler = hydra.utils.instantiate(
                scheduler_config,
                optimizer=self.optimizer,
                batch_size=1,
                len_loader=1,
            )
            self.add_callback(WarmupInit(scheduler_config=scheduler_config))
        else:
            self._scheduler = hydra.utils.instantiate(scheduler_config, optimizer=self.optimizer)

    def test(self) -> None:
        """Test the model."""
        if self.run_test and not self.config.trainer.get("fast_dev_run"):
            log.info("Starting testing!")
            log.info("Using last epoch's weights for testing.")
            self.trainer.test(datamodule=self.datamodule, model=self.module, ckpt_path=None)

    def export(self) -> None:
        """Deploy a model ready for production."""
        half_precision = "16" in self.trainer.precision

        input_shapes = self.config.export.input_shapes

        model_json, export_paths = export_model(
            config=self.config,
            model=self.module.model,
            export_folder=self.export_folder,
            half_precision=half_precision,
            input_shapes=input_shapes,
            idx_to_class=None,
        )

        if len(export_paths) == 0:
            return

        with open(os.path.join(self.export_folder, "model.json"), "w") as f:
            json.dump(model_json, f, cls=utils.HydraEncoder)


class Simsiam(SSL):
    """Simsiam model as a pytorch_lightning.LightningModule.

    Args:
        config: the main config
        checkpoint_path: if a checkpoint is specified, then it will return a trained model,
            with weights loaded from the checkpoint path specified.
            Defaults to None.
        run_test: Whether to run final test
    """

    def __init__(
        self,
        config: DictConfig,
        checkpoint_path: str | None = None,
        run_test: bool = False,
    ):
        super().__init__(config=config, checkpoint_path=checkpoint_path, run_test=run_test)
        self.backbone: nn.Module
        self.projection_mlp: nn.Module
        self.prediction_mlp: nn.Module

    def learnable_parameters(self) -> list[nn.Parameter]:
        """Get the learnable parameters."""
        return list(
            list(self.backbone.parameters())
            + list(self.projection_mlp.parameters())
            + list(self.prediction_mlp.parameters()),
        )

    def prepare(self) -> None:
        """Prepare the experiment."""
        super().prepare()
        self.backbone = hydra.utils.instantiate(self.config.model.model)
        self.projection_mlp = hydra.utils.instantiate(self.config.model.projection_mlp)
        self.prediction_mlp = hydra.utils.instantiate(self.config.model.prediction_mlp)
        self.optimizer = self.config.optimizer
        self.scheduler = self.config.scheduler
        self.module = self.config.model.module

    @property
    def module(self) -> LightningModule:
        """Get the module of the model."""
        return super().module

    @module.setter
    def module(self, module_config):
        """Set the module of the model."""
        module = hydra.utils.instantiate(
            module_config,
            model=self.backbone,
            projection_mlp=self.projection_mlp,
            prediction_mlp=self.prediction_mlp,
            optimizer=self.optimizer,
            lr_scheduler=self.scheduler,
        )
        if self.checkpoint_path is not None:
            module = module.__class__.load_from_checkpoint(
                self.checkpoint_path,
                model=self.backbone,
                projection_mlp=self.projection_mlp,
                prediction_mlp=self.prediction_mlp,
                criterion=module.criterion,
                optimizer=self.optimizer,
                lr_scheduler=self.scheduler,
            )
        self._module = module


class SimCLR(SSL):
    """SimCLR model as a pytorch_lightning.LightningModule.

    Args:
        config: the main config
        checkpoint_path: if a checkpoint is specified, then it will return a trained model,
            with weights loaded from the checkpoint path specified.
            Defaults to None.
        run_test: Whether to run final test
    """

    def __init__(
        self,
        config: DictConfig,
        checkpoint_path: str | None = None,
        run_test: bool = False,
    ):
        super().__init__(config=config, checkpoint_path=checkpoint_path, run_test=run_test)
        self.backbone: nn.Module
        self.projection_mlp: nn.Module

    def learnable_parameters(self) -> list[nn.Parameter]:
        """Get the learnable parameters."""
        return list(self.backbone.parameters()) + list(self.projection_mlp.parameters())

    def prepare(self) -> None:
        """Prepare the experiment."""
        super().prepare()
        self.backbone = hydra.utils.instantiate(self.config.model.model)
        self.projection_mlp = hydra.utils.instantiate(self.config.model.projection_mlp)
        self.optimizer = self.config.optimizer
        self.scheduler = self.config.scheduler
        self.module = self.config.model.module

    @property
    def module(self) -> LightningModule:
        return super().module

    @module.setter
    def module(self, module_config):
        """Set the module of the model."""
        module = hydra.utils.instantiate(
            module_config,
            model=self.backbone,
            projection_mlp=self.projection_mlp,
            optimizer=self.optimizer,
            lr_scheduler=self.scheduler,
        )
        if self.checkpoint_path is not None:
            module = module.__class__.load_from_checkpoint(
                self.checkpoint_path,
                model=self.backbone,
                projection_mlp=self.projection_mlp,
                criterion=module.criterion,
                optimizer=self.optimizer,
                lr_scheduler=self.scheduler,
            )
        self._module = module
        self._module.model = ModelSignatureWrapper(self._module.model)


class Barlow(SimCLR):
    """Barlow model as a pytorch_lightning.LightningModule.

    Args:
        config: the main config
        checkpoint_path: if a checkpoint is specified, then it will return a trained model,
            with weights loaded from the checkpoint path specified.
            Defaults to None.
        run_test: Whether to run final test
    """

    def __init__(
        self,
        config: DictConfig,
        checkpoint_path: str | None = None,
        run_test: bool = False,
    ):
        super().__init__(config=config, checkpoint_path=checkpoint_path, run_test=run_test)

    def prepare(self) -> None:
        """Prepare the experiment."""
        super(SimCLR, self).prepare()
        self.backbone = hydra.utils.instantiate(self.config.model.model)

        with open_dict(self.config):
            self.config.model.projection_mlp.hidden_dim = (
                self.config.model.projection_mlp.hidden_dim * self.config.model.projection_mlp_mult
            )
            self.config.model.projection_mlp.output_dim = (
                self.config.model.projection_mlp.output_dim * self.config.model.projection_mlp_mult
            )
        self.projection_mlp = hydra.utils.instantiate(self.config.model.projection_mlp)
        self.optimizer = self.config.optimizer
        self.scheduler = self.config.scheduler
        self.module = self.config.model.module


class BYOL(SSL):
    """BYOL model as a pytorch_lightning.LightningModule.

    Args:
        config: the main config
        checkpoint_path: if a checkpoint is specified, then it will return a trained model,
            with weights loaded from the checkpoint path specified.
            Defaults to None.
        run_test: Whether to run final test
        **kwargs: Keyword arguments
    """

    def __init__(
        self,
        config: DictConfig,
        checkpoint_path: str | None = None,
        run_test: bool = False,
        **kwargs: Any,
    ):
        super().__init__(
            config=config,
            checkpoint_path=checkpoint_path,
            run_test=run_test,
            **kwargs,
        )
        self.student_model: nn.Module
        self.teacher_model: nn.Module
        self.student_projection_mlp: nn.Module
        self.student_prediction_mlp: nn.Module
        self.teacher_projection_mlp: nn.Module

    def learnable_parameters(self) -> list[nn.Parameter]:
        """Get the learnable parameters."""
        return list(
            list(self.student_model.parameters())
            + list(self.student_projection_mlp.parameters())
            + list(self.student_prediction_mlp.parameters()),
        )

    def prepare(self) -> None:
        """Prepare the experiment."""
        super().prepare()
        self.student_model = hydra.utils.instantiate(self.config.model.student)
        self.teacher_model = hydra.utils.instantiate(self.config.model.student)
        self.student_projection_mlp = hydra.utils.instantiate(self.config.model.projection_mlp)
        self.student_prediction_mlp = hydra.utils.instantiate(self.config.model.prediction_mlp)
        self.teacher_projection_mlp = hydra.utils.instantiate(self.config.model.projection_mlp)
        self.optimizer = self.config.optimizer
        self.scheduler = self.config.scheduler
        self.module = self.config.model.module

    @property
    def module(self) -> LightningModule:
        return super().module

    @module.setter
    def module(self, module_config):
        """Set the module of the model."""
        module = hydra.utils.instantiate(
            module_config,
            student=self.student_model,
            teacher=self.teacher_model,
            student_projection_mlp=self.student_projection_mlp,
            student_prediction_mlp=self.student_prediction_mlp,
            teacher_projection_mlp=self.teacher_projection_mlp,
            optimizer=self.optimizer,
            lr_scheduler=self.scheduler,
        )
        if self.checkpoint_path is not None:
            module = module.__class__.load_from_checkpoint(
                self.checkpoint_path,
                student=self.student_model,
                teacher=self.teacher_model,
                student_projection_mlp=self.student_projection_mlp,
                student_prediction_mlp=self.student_prediction_mlp,
                teacher_projection_mlp=self.teacher_projection_mlp,
                criterion=module.criterion,
                optimizer=self.optimizer,
                lr_scheduler=self.scheduler,
            )
        self._module = module


class DINO(SSL):
    """DINO model as a pytorch_lightning.LightningModule.

    Args:
        config: the main config
        checkpoint_path: if a checkpoint is specified, then it will return a trained model,
            with weights loaded from the checkpoint path specified.
            Defaults to None.
        run_test: Whether to run final test
    """

    def __init__(
        self,
        config: DictConfig,
        checkpoint_path: str | None = None,
        run_test: bool = False,
    ):
        super().__init__(config=config, checkpoint_path=checkpoint_path, run_test=run_test)
        self.student_model: nn.Module
        self.teacher_model: nn.Module
        self.student_projection_mlp: nn.Module
        self.teacher_projection_mlp: nn.Module

    def learnable_parameters(self) -> list[nn.Parameter]:
        """Get the learnable parameters."""
        return list(
            list(self.student_model.parameters()) + list(self.student_projection_mlp.parameters()),
        )

    def prepare(self) -> None:
        """Prepare the experiment."""
        super().prepare()
        self.student_model = cast(nn.Module, hydra.utils.instantiate(self.config.model.student))
        self.teacher_model = cast(nn.Module, hydra.utils.instantiate(self.config.model.student))
        self.student_projection_mlp = cast(nn.Module, hydra.utils.instantiate(self.config.model.student_projection_mlp))
        self.teacher_projection_mlp = cast(nn.Module, hydra.utils.instantiate(self.config.model.teacher_projection_mlp))
        self.optimizer = self.config.optimizer
        self.scheduler = self.config.scheduler
        self.module = self.config.model.module

    @property
    def module(self) -> LightningModule:
        return super().module

    @module.setter
    def module(self, module_config):
        """Set the module of the model."""
        module = hydra.utils.instantiate(
            module_config,
            student=self.student_model,
            teacher=self.teacher_model,
            student_projection_mlp=self.student_projection_mlp,
            teacher_projection_mlp=self.teacher_projection_mlp,
            optimizer=self.optimizer,
            lr_scheduler=self.scheduler,
        )
        if self.checkpoint_path is not None:
            module = module.__class__.load_from_checkpoint(
                self.checkpoint_path,
                student=self.student_model,
                teacher=self.teacher_model,
                student_projection_mlp=self.student_projection_mlp,
                teacher_projection_mlp=self.teacher_projection_mlp,
                criterion=module.criterion,
                optimizer=self.optimizer,
                lr_scheduler=self.scheduler,
            )
        self._module = module


class EmbeddingVisualization(Task):
    """Visualization task for learned embeddings.

    Args:
        config: The loaded experiment config
        model_path: The path to a deployment model
        report_folder: Where to save the embeddings
        embedding_image_size: If not None rescale the images associated with the embeddings, tensorboard will save
            on disk a large sprite containing all the images in a matrix fashion, if the dimension of this sprite is too
            big it's not possible to load it in the browser. Rescaling the output image from the model input size to
            something smaller can solve this issue. The field is an int to always rescale to a squared image.
    """

    def __init__(
        self,
        config: DictConfig,
        model_path: str,
        report_folder: str = "embeddings",
        embedding_image_size: int | None = None,
    ):
        super().__init__(config=config)

        self.config = config
        self.metadata = {
            "report_files": [],
        }
        self.model_path = model_path
        self._device = utils.get_device()
        log.info("Current device: %s", self._device)

        self.report_folder = report_folder
        if self.model_path is None:
            raise ValueError(
                "Model path cannot be found!, please specify it in the config or pass it as an argument for evaluation"
            )
        self.embeddings_path = os.path.join(self.model_path, self.report_folder)
        if not os.path.exists(self.embeddings_path):
            os.makedirs(self.embeddings_path)
        self.embedding_writer = SummaryWriter(self.embeddings_path)
        self.writer_step = 0  # for tensorboard
        self.embedding_image_size = embedding_image_size
        self._deployment_model: BaseEvaluationModel
        self.deployment_model_type: str

    @property
    def device(self):
        return self._device

    @device.setter
    def device(self, device: str):
        self._device = device

        if self.deployment_model is not None:
            # After prepare
            self.deployment_model = self.deployment_model.to(self._device)

    @property
    def deployment_model(self):
        """Get the deployment model."""
        return self._deployment_model

    @deployment_model.setter
    def deployment_model(self, model_path: str):
        """Set the deployment model."""
        self._deployment_model = import_deployment_model(
            model_path=model_path, device=self.device, inference_config=self.config.inference
        )

    def prepare(self) -> None:
        """Prepare the evaluation."""
        super().prepare()
        self.deployment_model = self.model_path

    @torch.no_grad()
    def test(self) -> None:
        """Run embeddings extraction."""
        self.datamodule.setup("fit")
        idx_to_class = self.datamodule.val_dataset.idx_to_class
        self.datamodule.setup("test")
        dataloader = self.datamodule.test_dataloader()
        images = []
        metadata: list[tuple[int, str, str]] = []
        embeddings = []
        std = torch.tensor(self.config.transforms.std).view(1, -1, 1, 1)
        mean = torch.tensor(self.config.transforms.mean).view(1, -1, 1, 1)
        dl = self.datamodule.test_dataloader()
        counter = 0

        is_half_precision = False
        for param in self.deployment_model.parameters():
            if param.dtype == torch.half:
                is_half_precision = True
            break

        for batch in tqdm(dataloader):
            im, target = batch
            if is_half_precision:
                im = im.half()

            x = self.deployment_model(im.to(self.device))
            targets = [int(t.item()) for t in target]
            class_names = [idx_to_class[t.item()] for t in target]
            file_paths = [s[0] for s in dl.dataset.samples[counter : counter + len(im)]]
            embeddings.append(x.cpu())
            im = im * std
            im += mean

            if self.embedding_image_size is not None:
                im = interpolate(im, self.embedding_image_size)

            images.append(im.cpu())
            metadata.extend(zip(targets, class_names, file_paths, strict=False))
            counter += len(im)
        images = torch.cat(images, dim=0)
        embeddings = torch.cat(embeddings, dim=0)
        self.embedding_writer.add_embedding(
            embeddings,
            metadata=metadata,
            label_img=images,
            global_step=self.writer_step,
            metadata_header=["class", "class_name", "path"],
        )



from __future__ import annotations

import os
import shutil
from typing import Any

import cv2
from torchvision.datasets.mnist import MNIST

from quadra.datamodules import AnomalyDataModule
from quadra.utils.utils import get_logger

log = get_logger(__name__)


class MNISTAnomalyDataModule(AnomalyDataModule):
    """Standard anomaly datamodule with automatic download of the MNIST dataset."""

    def __init__(
        self, data_path: str, good_number: int, limit_data: int = 100, category: str | None = None, **kwargs: Any
    ):
        """Initialize the MNIST anomaly datamodule.

        Args:
            data_path: Path to the dataset
            good_number: Which number to use as a good class, all other numbers are considered anomalies.
            category: The category of the dataset. For mnist this is always None.
            limit_data: Limit the number of images to use for training and testing. Defaults to 100.
            **kwargs: Additional arguments to pass to the AnomalyDataModule.
        """
        super().__init__(data_path=data_path, category=None, **kwargs)
        self.good_number = good_number
        self.limit_data = limit_data

    def download_data(self) -> None:
        """Download the MNIST dataset and move images in the right folders."""
        log.info("Generating MNIST anomaly dataset for good number %s", self.good_number)

        mnist_train_dataset = MNIST(root=self.data_path, train=True, download=True)
        mnist_test_dataset = MNIST(root=self.data_path, train=False, download=True)

        self.data_path = os.path.join(self.data_path, "quadra_mnist_anomaly")

        if os.path.exists(self.data_path):
            shutil.rmtree(self.data_path)

        # Create the folder structure
        train_good_folder = os.path.join(self.data_path, "train", "good")
        test_good_folder = os.path.join(self.data_path, "test", "good")

        os.makedirs(train_good_folder, exist_ok=True)
        os.makedirs(test_good_folder, exist_ok=True)

        # Copy the good train images to the correct folder
        good_train_samples = mnist_train_dataset.data[mnist_train_dataset.targets == self.good_number]
        for i, image in enumerate(good_train_samples.numpy()):
            if i == self.limit_data:
                break
            cv2.imwrite(os.path.join(train_good_folder, f"{i}.png"), image)

        for number in range(10):
            if number == self.good_number:
                good_train_samples = mnist_test_dataset.data[mnist_test_dataset.targets == number]
                for i, image in enumerate(good_train_samples.numpy()):
                    if i == self.limit_data:
                        break
                    cv2.imwrite(os.path.join(test_good_folder, f"{number}_{i}.png"), image)
            else:
                test_bad_folder = os.path.join(self.data_path, "test", str(number))
                os.makedirs(test_bad_folder, exist_ok=True)
                bad_train_samples = mnist_train_dataset.data[mnist_train_dataset.targets == number]
                for i, image in enumerate(bad_train_samples.numpy()):
                    if i == self.limit_data:
                        break

                    cv2.imwrite(os.path.join(test_bad_folder, f"{number}_{i}.png"), image)

    def _prepare_data(self) -> None:
        """Prepare the MNIST dataset."""
        self.download_data()
        return super()._prepare_data()



from .barlowtwins import BarlowTwinsLoss
from .byol import BYOLRegressionLoss
from .dino import DinoDistillationLoss
from .idmm import IDMMLoss
from .simclr import SimCLRLoss
from .simsiam import SimSIAMLoss
from .vicreg import VICRegLoss

__all__ = [
    "BarlowTwinsLoss",
    "BYOLRegressionLoss",
    "IDMMLoss",
    "SimCLRLoss",
    "SimSIAMLoss",
    "VICRegLoss",
    "DinoDistillationLoss",
]



import pytest

from quadra.models.classification import TimmNetworkBuilder, TorchHubNetworkBuilder


@pytest.fixture
def resnet18():
    """Yield a resnet18 model."""
    yield TimmNetworkBuilder("resnet18", pretrained=False, freeze=True, exportable=True)


@pytest.fixture
def resnet50():
    """Yield a resnet50 model."""
    yield TimmNetworkBuilder("resnet50", pretrained=False, freeze=True, exportable=True)


@pytest.fixture
def vit_tiny_patch16_224():
    """Yield a vit_tiny_patch16_224 model."""
    yield TimmNetworkBuilder("vit_tiny_patch16_224", pretrained=False, freeze=True, exportable=True)


@pytest.fixture
def dino_vits8():
    """Yield a dino_vits8 model."""
    yield TorchHubNetworkBuilder(
        repo_or_dir="facebookresearch/dino:main",
        model_name="dino_vits8",
        pretrained=False,
        freeze=True,
        exportable=True,
    )


@pytest.fixture
def dino_vitb8():
    """Yield a dino_vitb8 model."""
    yield TorchHubNetworkBuilder(
        repo_or_dir="facebookresearch/dino:main",
        model_name="dino_vitb8",
        pretrained=False,
        freeze=True,
        exportable=True,
    )
